import logging

from ploomber.tasks.abc import Task
from ploomber.sources import FileSource
from ploomber.util import requires


class UploadToS3(Task):
    """
    Upload a file to S3

    Parameters
    ----------
    source: str
        Path to file to upload
    product: ploomber.products.genericproduct
        Product must be initialized with the desired Key
    dag: ploomber.DAG
        A DAG to add this task to
    name: str
        A str to indentify this task. Should not already exist in the dag
    bucket: str
        Bucked to upload
    """
    def __init__(self,
                 source,
                 product,
                 dag,
                 bucket,
                 name=None,
                 params=None,
                 client_kwargs=None,
                 upload_file_kwargs=None):
        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = type(self)._init_source(source, kwargs)
        super().__init__(product, dag, name, params)
        self._bucket = bucket
        self._client_kwargs = client_kwargs
        self._upload_file_kwargs = upload_file_kwargs

    @requires(['boto3'], 'UploadToS3')
    def run(self):
        import boto3
        from botocore.exceptions import ClientError

        client_kwargs = self._client_kwargs or {}
        upload_file_kwargs = self._upload_file_kwargs or {}
        s3_client = boto3.client('s3', **client_kwargs)
        source = str(self.source)

        try:
            s3_client.upload_file(source, self._bucket, str(self.product),
                                  upload_file_kwargs)
        except ClientError as e:
            logging.error(e)

    @staticmethod
    def _init_source(source, kwargs):
        return FileSource(source, **kwargs)



from ploomber.codediffer import CodeDiffer


def _logging_factory():
    return None


class DAGConfiguration:
    """
    DAGConfiguration() initializes a configuration object with default values.
    """
    __attrs = {
        'outdated_by_code', 'cache_rendered_status', 'logging_factory',
        'differ', 'hot_reload'
    }

    @classmethod
    def from_dict(cls, d):
        cfg = cls()

        for key, value in d.items():
            setattr(cfg, key, value)

        return cfg

    def __init__(self):
        self.outdated_by_code = True
        self.cache_rendered_status = False
        self.hot_reload = False
        self.logging_factory = _logging_factory
        self.differ = CodeDiffer()

    @property
    def outdated_by_code(self):
        return self._outdated_by_code

    @outdated_by_code.setter
    def outdated_by_code(self, value):
        if value not in {True, False}:
            raise ValueError('outdated_by_code must be True or False')

        self._outdated_by_code = value

    @property
    def hot_reload(self):
        return self._hot_reload

    @hot_reload.setter
    def hot_reload(self, value):
        if value not in {True, False}:
            raise ValueError('hot_reload must be True or False')

        self._hot_reload = value

    @property
    def cache_rendered_status(self):
        return self._cache_rendered_status

    @cache_rendered_status.setter
    def cache_rendered_status(self, value):
        if value not in {True, False}:
            raise ValueError('cache_rendered_status must be True or False')

        self._cache_rendered_status = value

    @property
    def differ(self):
        return self._differ

    @differ.setter
    def differ(self, value):
        if value == 'default':
            value = CodeDiffer()

        self._differ = value

    @property
    def logging_factory(self):
        return self._logging_factory

    @logging_factory.setter
    def logging_factory(self, value):
        self._logging_factory = value

    def __setattr__(self, key, value):
        # prevent non-existing parameters from being created
        attr_name = key[1:] if key.startswith('_') else key
        if attr_name not in self.__attrs:
            raise AttributeError('"{}" is not a valid parameter, must be '
                                 'one of: {}'.format(key, self.__attrs))
        super().__setattr__(key, value)



# NOTE: should this be part of each Source?
import warnings

from ploomber.util import requires


class DAGQualityChecker:

    def __call__(self, dag):
        self.empty_docstrings(dag)

    def empty_docstrings(self, dag):
        """Evaluate code quality
        """
        for name, task in dag.items():

            doc = task.source.doc

            if doc is None or doc == '':
                warnings.warn('Task "{}" has no docstring'.format(name))


@requires(['numpydoc'])
def diagnose(source):
    """Prints some diagnostics
    """
    from numpydoc.docscrape import NumpyDocString

    # [WIP] function to validate docstrings in sources that
    # have placeholders
    found = source.value.variables
    docstring_np = NumpyDocString(source.value.docstring())
    documented = set([p[0] for p in docstring_np['Parameters']])

    print('The following variables were found in the template but are '
          'not documented: {}'.format(found - documented))

    print('The following variables are documented but were not found in '
          'the template: {}'.format(documented - found))

    return documented, found



"""
Utlity functions to run Ploomber tests. This is not a public module
"""


def assert_no_extra_attributes_in_class(abstract_class,
                                        concrete_class,
                                        allowed=None):
    """
    Ploomber makes heavy use of abstract classes to provide a uniform API for
    tasks, products, metadata, etc. When defining abstract classes, the
    interpreter refuses to instantiate an object where the concrete class
    misses implementation for abstract methods. However, it does not complain
    if the concrete class implements *extra methods*

    This has been problematic to remove old code. As we simplify the API,
    sometimes concrete classes become outdated. For example, before creating
    Metadata, all metadata logic was embedded in the Product objects, when
    we made the change, we removed some abstract methods from the Product class
    but it took a long time to realize that we sould've removed these mehods
    from the MetaProduct class as well. We use this function to alert us
    when there are things we can remove.

    The other case also happens: we add functionality to concrete classes but
    we do not do it in the abstract class, when this happens we have to decide
    whether to add them to the abstract class (recommended) or make an
    exception in such case, those new methods should be named with double
    leading underscore to be ignored by this check and to prevent polluting the
    public interface. Single leading underscore methods are checked to allow
    abstract classes define its own private API, which is also important
    for consistency, even if the end user does not use these methods.

    Convention:
        - no leading underscode: public API
        - one leading underscore: private Ploomber API. Not meant to be used by
            end-users but can be user by developers
        - two leading underscores: private class API. Not meant to be used
            outside the implementation of the class itself. Abstract classes
            should not define these, these are intended to carry logic
            specific to concrete classes

    NOTE: maybe a better alternative to allowed is to create an abstract
    class that adds new abstract methods
    """
    allowed = allowed or set()

    # allow "private" methods
    preffixes = [
        '_{}__'.format(class_.__name__) for class_ in concrete_class.__bases__
    ] + ['__', '_', '_{}__'.format(concrete_class.__name__)]

    extra_attrs = {
        attr
        for attr in set(dir(concrete_class)) - set(dir(abstract_class))
        if not any(attr.startswith(p) for p in preffixes)
    } - allowed

    if extra_attrs:
        raise ValueError('The following methods/attributes in {} '
                         'are not part of the {} interface: {}'.format(
                             concrete_class.__name__, abstract_class.__name__,
                             extra_attrs))



"""
A client reflects a connection to a system that performs the actual
computations
"""
import abc
import logging


class Client(abc.ABC):
    """
    Abstract class for all clients

    Clients are classes that communicate with another system (usually a
    database), they provide a thin wrapper around libraries that implement
    clients to avoid managing connections directly. The most common use
    case by far is for a Task/Product to submit some code to a system,
    a client just provides a way of doing so without dealing with connection
    details.

    A Client is reponsible for making sure an open connection is available
    at any point (open a connection if none is available).

    However, clients are not strictly necessary, a Task/Product could manage
    their own client connections. For example the NotebookRunner task does have
    a Client since it only calls an external library to run.


    Notes
    -----
    Method's names were chosen to resemble the ones in the Python DB API Spec
    2.0 (PEP 249)

    """
    def __init__(self):
        self._set_logger()

    @property
    @abc.abstractmethod
    def connection(self):
        """Return a connection, open one if there isn't any
        """
        pass

    @abc.abstractmethod
    def execute(self, code):
        """Execute code
        """
        pass

    @abc.abstractmethod
    def close(self):
        """Close connection if there is one active
        """
        pass

    # __getstate__ and __setstate__ are needed to make this picklable

    def __getstate__(self):
        state = self.__dict__.copy()
        # _logger is not pickable, so we remove them and build it
        # again in __setstate__
        del state['_logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._set_logger()

    def _set_logger(self):
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))



"""
Metadata represents the information we need to save in order to support
incremental builds: source code and build timestmp
"""
import logging
import warnings
import abc
from datetime import datetime
from copy import deepcopy

from ploomber.util.util import callback_check
from ploomber.products._resources import process_resources
from ploomber.products.serializeparams import remove_non_serializable_top_keys


class AbstractMetadata(abc.ABC):
    """Abstract class to represent Product's metadata

    If product does not exist, initialize empty metadata, otherwise use
    ``product.fetch_metadata`` to load it
    """
    def __init__(self, product):
        self.__data = None
        self._product = product

        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))

    @property
    @abc.abstractmethod
    def _data(self):
        """
        Private API, returns the dictionary representation of the metadata
        """
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def timestamp(self):
        """When the product was originally created
        """
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def stored_source_code(self):
        """Source code that generated the product
        """
        pass  # pragma: no cover

    @property
    @abc.abstractclassmethod
    def params(self):
        """Task params
        """
        pass

    @abc.abstractmethod
    def update(self, source_code, params):
        """
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def delete(self):
        """Delete metadata
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _get(self):
        """Load metadata
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def clear(self):
        """
        Clear tne in-memory copy, if the metadata is accessed again, it should
        trigger another call to load()
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def update_locally(self, data):
        """
        Updates metadata locally. Called when tasks are successfully
        executed in a subproces, to sync the values in the main
        process (because the call to .update() happens in the subprocess as
        well)
        """
        pass  # pragma: no cover

    def to_dict(self):
        """Returns a dict copy of ._data
        """
        return deepcopy(self._data)

    def __eq__(self, other):
        self._fetch()

        if isinstance(other, type(self)):
            # metadata is lazily loaded, ensure you have a local copy
            other._fetch()
            return self._data == other._data
        else:
            return self._data == other

    def __getstate__(self):
        state = self.__dict__.copy()

        if '_logger' in state:
            del state['_logger']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))

    def _fetch(self):
        """Fetches metadata if needed. If already fetched, does nothing
        """
        if not self._did_fetch:
            self._get()


class Metadata(AbstractMetadata):
    """
    Internal class to standardize access to Product's metadata

    This implementation tries to avoid fetching metadata when it can, because
    it might be a slow process. Since this class also performs metadata
    validation, there are cases when the metadata here and the one in the
    storage backend won't match, for example when the product does not exist,
    metadata in the storage backend is simply ignored. The values returned
    by this class should be considered the "true metadata".

    Note that Metadata relies the actual Product implementation for
    most operations. The objective of this class is to encapsuate logic
    that requires calling several of Products's methods and do validation
    upon metadata fetching.

    Attributes
    ----------
    timestamp
        Last updated product timestamp
    stored_source_code
        Last updates product source code
    """
    def _default_metadata(self):
        return dict(timestamp=None, stored_source_code=None, params=None)

    def __init__(self, product):
        self.__data = None
        self._product = product

        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))
        self._did_fetch = False

    @property
    def params(self):
        return self._data.get('params')

    @property
    def timestamp(self):
        return self._data.get('timestamp')

    @property
    def stored_source_code(self):
        return self._data.get('stored_source_code')

    @property
    def _data(self):
        if not self._did_fetch:
            self._get()

        return self.__data

    @_data.setter
    def _data(self, value):
        # if data changes, cached status in product is no longer valid
        self._product._reset_cached_outdated_status()
        self.__data = value

    def _get(self):
        """
        Get the "true metadata", ignores actual metadata if the product does
        not exist. It is lazily called by the timestamp and stored_source_code
        attributes. Ignores fetched metadata and replaces it with the stored
        metadata
        """
        # if the product does not exist, ignore metadata in backend storage

        # FIXME: cache the output of this command, we are using it in several
        # places, sometimes we have to re-fetch but sometimes we can cache,
        # look for product.exists() references and .exists() references
        # in the Product definition
        if not self._product.exists():
            metadata = self._default_metadata()
        else:
            # FIXME: if anything goes wrong when fetching metadata, warn
            # and set it to a valid dictionary with None values, validation
            # should happen here, not in the fetch_metadata method, but we
            # shouldn't catch all exceptions. Create a new one in
            # ploomber.exceptions and raise it on each fetch_metadata
            # implementation when failing to unserialize
            metadata_fetched = self._product.fetch_metadata()

            if metadata_fetched is None:
                self._logger.debug(
                    'fetch_metadata for product %s returned '
                    'None', self._product)
                metadata = self._default_metadata()
            else:
                # FIXME: we need to further validate this, need to check
                # that this is an instance of mapping, if yes, then
                # check keys [timestamp, stored_source_code], check
                # types and fill with None if any of the keys is missing
                metadata = metadata_fetched

        self._did_fetch = True
        self._data = metadata

    def update(self, source_code, params):
        """
        Update metadata in the storage backend, this should be called by
        Task objects when running successfully to update metadata in the
        backend storage. If saving in the backend storage succeeds the local
        copy is updated as well

        Parameters
        ----------
        source_code : str
            Task's source code

        params : dict
            Task's params
        """
        # remove any unserializable parameters
        params = remove_non_serializable_top_keys(params)

        new_data = dict(
            timestamp=datetime.now().timestamp(),
            stored_source_code=source_code,
            # process params to store hashes in case they're
            # declared as resources
            params=process_resources(params))

        kwargs = callback_check(self._product.prepare_metadata,
                                available={
                                    'metadata': new_data,
                                    'product': self._product
                                })

        data = self._product.prepare_metadata(**kwargs)

        self._product.save_metadata(data)

        # if saving worked, we can update the local in-memory copy
        self.update_locally(new_data)

    def update_locally(self, data):
        """Updates the in-memory copy, does not update persistent copy
        """
        # could be the case that we haven't fetched metadata yet. since this
        # overwrites existing metadata. we no longer have to fetch
        self._did_fetch = True
        self._data = deepcopy(data)

    def delete(self):
        """Calls ._product._delete_metadata()
        """
        self._product._delete_metadata()
        self._data = self._default_metadata()

    def clear(self):
        """
        Clears up metadata local copy, next time the timestamp or
        stored_source_code are needed, this will trigger another call to
        ._get(). Should be called only when the local copy might be outdated
        due external execution. Currently, we are only using this when running
        DAG.build_partially, because that triggers a deep copy of the original
        DAG. hence our local copy in the original DAG is not valid anymore
        """
        self._did_fetch = False
        self._data = self._default_metadata()

    def __repr__(self):
        return f'{type(self).__name__}({self._data!r})'


class MetadataCollection(AbstractMetadata):
    """Metadata class used for MetaProduct
    """

    # FIXME: this can be optimized. instead of keeping separate copies for each
    # the metadata objects can share the underlying dictionary, since they
    # must have the same values anyway, this allows to remove the looping logic
    def __init__(self, products):
        self._products = products

    @property
    def timestamp(self):
        timestamps = [
            p.metadata.timestamp for p in self._products
            if p.metadata.timestamp is not None
        ]

        any_none = any(p.metadata.timestamp is None for p in self._products)

        # edge cases: 1) any of the timestamps is none, this is mostly
        # due to metadata corruption, we can no longer compute the timestamp
        # reliable. 2) no timestamps at all happens when the task hasn't been
        # executed
        if any_none or not timestamps:
            # warn on corrupted data
            if any_none and timestamps:
                warnings.warn(f'Corrupted product metadata ({self!r}): '
                              'at least one product had a null timestamp, '
                              'but others had non-null timestamp')

            return None
        else:
            # timestamps should usually be very close, but there can be
            # differences for some products whose metadata takes some time
            # to save (e.g. remote db) in such case, we use the
            # minimum to cover the edge case where another process runs an
            # upstream dependency in between saving metadata
            return min(timestamps)

    @property
    def stored_source_code(self):
        stored_source_code = [
            p.metadata.stored_source_code for p in self._products
        ]
        # if source code differs (i.e. more than one element)
        if len(set(stored_source_code)) > 1:
            warnings.warn(
                'Stored source codes for products {} '
                'are different, but they are part of the same '
                'MetaProduct, returning stored_source_code as None'.format(
                    self._products))
            return None
        else:
            return stored_source_code[0]

    @property
    def params(self):
        return self._products.first.params

    def update(self, source_code, params):
        for p in self._products:
            p.metadata.update(source_code, params)

    def update_locally(self, data):
        for p in self._products:
            p.metadata.update_locally(data)

    def delete(self):
        for p in self._products:
            p.metadata.delete()

    def _get(self):
        for p in self._products:
            p.metadata._get()

    def clear(self):
        for p in self._products:
            p.metadata.clear()

    def to_dict(self):
        products = list(self._products)
        source = set(p.metadata.stored_source_code for p in products)
        large_diff = large_timestamp_difference(p.metadata.timestamp
                                                for p in products)

        # warn if metadata does not match, give a little tolerance (5 seconds)
        # for timestamps since they are expected to have slight differences
        if len(source) > 1 or large_diff:
            warnings.warn(f'Metadata acros products ({self!r}) differs, '
                          'this could be due to metadata corruption or '
                          'slow metadata storage backend, returning the '
                          'metadata from the first product')

        return products[0].metadata.to_dict()

    @property
    def _data(self):
        products = list(self._products)
        source = set(p.metadata.stored_source_code for p in products)
        large_diff = large_timestamp_difference(p.metadata.timestamp
                                                for p in products)

        # warn if metadata does not match, give a little tolerance (5 seconds)
        # for timestamps since they are expected to have slight differences
        if len(source) > 1 or large_diff:
            warnings.warn(f'Metadata acros products ({self!r}) differs, '
                          'this could be due to metadata corruption or '
                          'slow metadata storage backend, returning the '
                          'metadata from the first product')

        return products[0].metadata._data


def large_timestamp_difference(timestamps):
    """Returns True if there is at least one timestamp difference > 5 seconds
    """
    dts = [datetime.fromtimestamp(ts) for ts in timestamps]

    for i in range(len(dts)):
        for j in range(len(dts)):
            if i != j:
                diff = (dts[i] - dts[j]).total_seconds()

                if abs(diff) > 5:
                    return True

    return False


class MetadataAlwaysUpToDate(AbstractMetadata):
    """
    Metadata for Link tasks (always up-to-date)
    """
    def __init__(self):
        pass

    @property
    def timestamp(self):
        return 0

    @property
    def stored_source_code(self):
        return None

    @property
    def params(self):
        return None

    def _get(self):
        pass  # pragma: no cover

    def update(self, source_code, params):
        pass  # pragma: no cover

    def update_locally(self, data):
        pass  # pragma: no cover

    @property
    def _data(self):
        return {'timestamp': 0, 'stored_source_code': None, 'params': {}}

    def delete(self):
        pass  # pragma: no cover

    def clear(self):
        pass  # pragma: no cover



import string
from ploomber.static_analysis.parser.tokens import (Integer, BinaryOperator,
                                                    Assignment, Name, Operator,
                                                    String, Null)
from ploomber.static_analysis.parser.abstract import Lexer


class RLexer(Lexer):
    """
    Parses a code string and returns one token at a time

    Notes
    -----
    R data structures: http://adv-r.had.co.nz/Data-structures.html
    """
    def __init__(self, text):
        self.text = text
        self.pos = 0
        self.current_char = self.text[0]

    @property
    def next_char(self):
        return self.text[self.pos + 1]

    def comes_next(self, s):
        return self.text[self.pos:self.pos + len(s)] == s

    def advance(self, n=1):
        advanced = self.text[self.pos:self.pos + n]

        self.pos += n

        if self.pos > len(self.text) - 1:
            self.current_char = None
        else:
            self.current_char = self.text[self.pos]

        return advanced

    def skip_whitespace(self):
        while self.current_char is not None and self.current_char.isspace():
            self.advance()

    def read_name(self):
        result = ''

        while (self.current_char is not None
               and self.current_char in string.ascii_letters):
            result += self.current_char
            self.advance()

        return result

    def read_string(self):
        self.advance()

        result = ''

        while (self.current_char is not None
               and self.current_char not in ['"', "'"]):
            result += self.current_char
            self.advance()

        self.advance()

        return result

    def read_integer(self):
        result = ''

        while self.current_char is not None and self.current_char.isdigit():
            result += self.current_char
            self.advance()

        return int(result)

    def __iter__(self):
        while self.current_char is not None:

            if self.current_char.isspace():
                self.skip_whitespace()
                continue

            elif self.comes_next('list'):
                op = Operator('list')
                self.advance(n=len('list'))
                yield op

            elif self.comes_next('NULL'):
                null = Null()
                self.advance(n=len('NULL'))
                yield null

            # Vector definiton start
            elif self.current_char == 'c' and self.next_char == '(':
                self.advance(n=2)
                yield Operator('c(')

            elif self.current_char in string.ascii_letters:
                yield Name(self.read_name())

            # Assignment (<-)
            elif self.current_char == '<' and self.next_char == '-':
                self.advance(n=2)
                yield Assignment('<-')

            elif self.current_char == '=':
                self.advance()
                yield Assignment('=')

            # Other operators
            elif self.current_char in ['(', ')', ',']:
                op = Operator(self.current_char)
                self.advance()
                yield op

            elif self.current_char in ['"', "'"]:
                yield String(self.read_string())

            elif self.current_char.isdigit():
                yield Integer(self.read_integer())

            elif self.current_char in ['+', '-', '*']:
                op = BinaryOperator(self.current_char)
                self.advance()
                yield op
            else:
                raise SyntaxError('Could not parse parameters in R notebook. '
                                  'Verify they have the right format: '
                                  '"upstream = list(\'a\', \'b\')" or '
                                  '"product = list(name=\'path/to/file\')"')



from pathlib import Path
import os
import logging
import sys
import importlib
import inspect
import argparse
from collections.abc import Mapping
import warnings
try:
    import importlib.resources as importlib_resources
except ImportError:  # pragma: no cover
    # backported
    import importlib_resources

from ploomber.spec.dagspec import DAGSpec
from ploomber.env.envdict import EnvDict
from ploomber.util.dotted_path import load_dotted_path
from ploomber.util import default
from ploomber.entrypoint import EntryPoint


def process_arg(s):
    clean = None

    if s.startswith('--'):
        clean = s[2:]
    elif s.startswith('-'):
        clean = s[1:]
    else:
        clean = s

    return clean.replace('-', '_')


class CustomMutuallyExclusiveGroup(argparse._MutuallyExclusiveGroup):
    """
    A subclass of argparse._MutuallyExclusiveGroup that determines whether
    the added args should go into the static of dynamic API
    """
    def add_argument(self, *args, **kwargs):
        if not self._container.finished_static_api:
            if (not self._container.in_context
                    and self._container.finished_init):
                raise RuntimeError('Cannot add arguments until the static '
                                   'API has been declared')
            else:
                # running inside the context manager
                self._container.static_args.extend(
                    [process_arg(arg) for arg in args])

        # outside context manager
        return super().add_argument(*args, **kwargs)


class CustomParser(argparse.ArgumentParser):
    """
    Most of our CLI commands operate on entry points, the CLI signature is
    defined by a few static arguments (they always appear when doing
    ploomber {cmd} --entry-point {entry} --help) and dynamic arguments
    (they are generated depending on the entry point). To support this, we
    have this CustomParser that distinguishes between static and dynamic args

    Once initialized, static args must be added using this as a context
    manager, only after this has happened, other arguments can be added using
    parser.add_argument.
    """
    def __init__(self, *args, **kwargs):
        self.DEFAULT_ENTRY_POINT = default.try_to_find_entry_point()

        self.static_args = []
        self.finished_static_api = False
        self.in_context = False
        self.finished_init = False
        super().__init__(*args, **kwargs)

        self.add_argument('--log',
                          '-l',
                          help='Enables logging to stdout at the '
                          'specified level',
                          default=None)

        self.add_argument('--log-file',
                          '-F',
                          help='Enables logging to the given file',
                          default=None)

        if self.DEFAULT_ENTRY_POINT:
            entry_point_help = ('Entry point, defaults '
                                f'to {self.DEFAULT_ENTRY_POINT}')
            if os.environ.get('ENTRY_POINT'):
                entry_point_help += ' (ENTRY_POINT env var)'
        else:
            entry_point_help = 'Entry point'

        self.add_argument('--entry-point',
                          '-e',
                          help=entry_point_help,
                          default=self.DEFAULT_ENTRY_POINT)

        self.finished_init = True

    def parse_entry_point_value(self):
        """
        Returns the entry_point value pased without calling parse_args(),
        this is required to find env params to show, if we call parse_args()
        the CLI stops there and shows available params
        """
        index = None

        try:
            index = sys.argv.index('--entry-point')
        except ValueError:
            pass

        try:
            index = sys.argv.index('-e')
        except ValueError:
            pass

        # no --entry-point/-e arg passed, use default
        if index is None:
            if self.DEFAULT_ENTRY_POINT is None:
                self.error('Unable to find a pipeline. '
                           'Use --entry-point/-e to pass a '
                           'entry point\'s location or '
                           'place it in a standard location.\n\n'
                           'Need help? https://ploomber.io/community')

            return self.DEFAULT_ENTRY_POINT
        else:
            try:
                return sys.argv[index + 1]
            except IndexError:
                pass

            # replicate the original message emitted by argparse
            action = self._option_string_actions['-e']
            options = '/'.join(action.option_strings)
            self.error(f'argument {options}: expected one argument')

    def add_argument(self, *args, **kwargs):
        """
        Add a CLI argument. If called after the context manager, it is
        considered part of the dynamic API, if called within the context
        manager, the arg is considered part of the static API. If it's
        called outside a context manager, and no static API has been set,
        it raises an error
        """
        if not self.finished_static_api:
            if not self.in_context and self.finished_init:
                raise RuntimeError('Cannot add arguments until the static '
                                   'API has been declared')
            else:
                # running inside the context manager
                self.static_args.extend([process_arg(arg) for arg in args])

        # outside context manager
        return super().add_argument(*args, **kwargs)

    def add_mutually_exclusive_group(self, **kwargs):
        """
        Add a mutually exclusive group. It returns a custom class that
        correctly stores the arguments in the static or dynamic API
        """
        group = CustomMutuallyExclusiveGroup(self, **kwargs)
        self._mutually_exclusive_groups.append(group)
        return group

    def __enter__(self):
        self.in_context = True
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.in_context = False
        self.finished_static_api = True

    def process_factory_dotted_path(self, dotted_path):
        """Parse a factory entry point, returns initialized dag and parsed args
        """
        entry = load_dotted_path(str(dotted_path), raise_=True)

        # add args using the function's signature
        required, _ = _add_args_from_callable(self, entry)

        # if entry point was decorated with @with_env, add arguments
        # to replace declared variables in env.yaml
        if hasattr(entry, '_env_dict'):
            _add_cli_args_from_env_dict_keys(self, entry._env_dict)

        args = self.parse_args()

        _configure_logger(args)

        # extract required (by using function signature) params from the cli
        # args
        kwargs = {key: getattr(args, key) for key in required}

        # env and function defaults replaced
        replaced = _env_keys_to_override(args, self.static_args)

        # TODO: add a way of test this by the parameters it will use to
        # call the function, have an aux function to get those then another
        # to execute, test using the first one
        dag = entry(**{**kwargs, **replaced})

        return dag, args

    def load_from_entry_point_arg(self):
        """Parses an entry point, adding arguments by extracting them from the env.

        Returns a dag and the parsed args
        """
        entry_point = EntryPoint(self.parse_entry_point_value())
        dag, args = load_dag_from_entry_point_and_parser(
            entry_point, self, sys.argv)
        return dag, args


def _path_for_module_path(module_path):
    mod_name, path_part = module_path.split('::')

    # TODO: check it's only two parts after splitting
    with importlib_resources.path(mod_name, path_part) as p:
        path = str(p)

    return path


def _first_non_empty_line(doc):
    for line in doc.split('\n'):
        if line:
            return line.strip()


def _parse_doc(callable_):
    """
    Convert numpydoc docstring to a list of dictionaries
    """
    doc = callable_.__doc__

    # no docstring
    if doc is None:
        return {'params': {}, 'summary': None}

    # try to import numpydoc, if can't find it, just returnt the first line
    try:
        docscrape = importlib.import_module('numpydoc.docscrape')
    except ModuleNotFoundError:
        return {'params': {}, 'summary': _first_non_empty_line(doc)}

    doc_parsed = docscrape.NumpyDocString(doc)

    parameters = {
        p.name: {
            'desc': ' '.join(p.desc),
            'type': p.type
        }
        for p in doc_parsed['Parameters']
    }

    # docscrape returns one element per line
    summary = 'Docstring: {}'.format('\n'.join(doc_parsed['Summary']))

    return {'params': parameters, 'summary': summary}


def _env_keys_to_override(args, static_args):
    """
    Returns a dictionary with all extra cli parameters passed, all these must
    be parameters that part of the env or params (with no defaults) if
    entry point is a factory function
    """
    return {
        name: getattr(args, name)
        for name in dir(args) if not name.startswith('_')
        if getattr(args, name) is not None if name not in static_args
    }


def _add_cli_args_from_env_dict_keys(parser, env_dict):
    """
    Add one parameter to the args parser by taking a look at all values
    defined in an env dict object
    """
    # flatten keys from the env dictionary. e.g. from {'a': {'b': 1}} is
    # converted to {'a--b': 1}. This allows us to add cli args such as --a--b
    # to modify any key in the env. Note that we use double hyphens to have
    # an unambious section separator. Environments are commonly loaded from
    # YAML files. Keys in such files might contain hyphens/underscores, we
    # allow users to have those characters but double hyphens/underscores are
    # not permitted as they'd conflict with the CLI generation logic
    flat_env_dict = _flatten_dict(env_dict._data)

    for arg, val in flat_env_dict.items():
        # do not add default keys like {{cwd}}, {{here}}
        if arg not in env_dict.default_keys:
            parser.add_argument('--env--' + arg,
                                help='Default: {}'.format(val))


def _parse_signature_from_callable(callable_):
    """
    Parse a callable signature, return a dictionary with
    {param_key: default_value} and a list of required parameters
    """
    sig = inspect.signature(callable_)

    required = [
        k for k, v in sig.parameters.items() if v.default == inspect._empty
    ]

    defaults = {
        k: v.default
        for k, v in sig.parameters.items() if v.default != inspect._empty
    }

    return required, defaults, sig.parameters


def get_desc(doc, arg):
    arg_data = doc['params'].get(arg)
    return None if arg_data is None else arg_data['desc']


def add_argument_kwargs(params, arg):
    """
    Build kwargs for parser.add_argument function
    """
    valid_hints = [int, float, str, bool]

    fn_default = params[arg].default
    fn_annotation = params[arg].annotation

    # special case, bool with default value becomes a flag
    if fn_annotation is bool and fn_default is not inspect._empty:
        kwargs = {'action': 'store_true' if not fn_default else 'store_false'}
    elif fn_annotation in valid_hints:
        kwargs = {'type': fn_annotation, 'default': fn_default}
    else:
        kwargs = {'default': fn_default}

    return kwargs


def _add_args_from_callable(parser, callable_):
    """
    Modifies an args parser to include parameters from a callable, adding
    parameters with default values as optional and parameters with no defaults
    as mandatory. Adds descriptions from parsing the callable's docstring

    It also adds the description from the docstring, if any

    Returns parsed args: required (list) and defaults (dict)
    """
    doc = _parse_doc(callable_)
    required, defaults, params = _parse_signature_from_callable(callable_)

    for arg in defaults.keys():
        conflict = False

        try:
            parser.add_argument('--' + arg,
                                help=get_desc(doc, arg),
                                **add_argument_kwargs(params, arg))
        except argparse.ArgumentError as e:
            conflict = e

        if conflict:
            if 'conflicting option string' in conflict.message:
                raise ValueError(
                    f'The signature from {callable_.__name__!r} '
                    'conflicts with existing arguments in the command-line '
                    'interface, please rename the following '
                    f'argument: {arg!r}')
            else:
                raise conflict

    for arg in required:
        parser.add_argument(arg,
                            help=get_desc(doc, arg),
                            **add_argument_kwargs(params, arg))

    if doc['summary']:
        desc = parser.description
        parser.description = '{}. {}'.format(desc, doc['summary'])

    return required, defaults


def _process_file_dir_or_glob(parser, dagspec_arg=None):
    """
    Process a file entry point file, directory or glob-like pattern,
    the initialized dag and parsed args

    Parameters
    ----------
    parser : CustomParser
        CLI arg parser
    """
    # NOTE: we must use parser.parse_entry_point_value() instead or
    # args.parse_args because calling the latter wont allow us to add more
    # cli parameters, but we want that to expose parms from env
    entry_point_value = dagspec_arg or parser.parse_entry_point_value()
    entry = EntryPoint(entry_point_value)

    if entry.type in {EntryPoint.Directory, EntryPoint.Pattern}:
        # pipelines initialized from directories or patterns cannot be
        # parametrized
        path_to_env = None
    # file
    else:
        path_to_env = default.path_to_env_from_spec(entry_point_value)

    if path_to_env:
        env_dict = EnvDict(path_to_env,
                           path_to_here=Path(entry_point_value).parent
                           if entry.type == EntryPoint.File else None)
        _add_cli_args_from_env_dict_keys(parser, env_dict)

    args = parser.parse_args()
    dagspec_arg = dagspec_arg or args.entry_point

    _configure_logger(args)

    entry_point = EntryPoint(dagspec_arg)

    # directory
    if entry_point.type == EntryPoint.Directory:
        dag = DAGSpec.from_directory(dagspec_arg).to_dag()
    # pattern
    elif entry_point.type == EntryPoint.Pattern:
        dag = DAGSpec.from_files(dagspec_arg).to_dag()
    # file
    else:
        if path_to_env:
            # and replace keys depending on passed cli args
            replaced = _env_keys_to_override(args, parser.static_args)
            env = env_dict._replace_flatten_keys(replaced)
            dag = DAGSpec(dagspec_arg, env=env).to_dag()
        else:
            dag = DAGSpec(dagspec_arg).to_dag()

    return dag, args


def load_dag_from_entry_point_and_parser(entry_point, parser, argv):
    """Load DAG from entry point

    Parameters
    ----------
    parser : CustomParser
        The cli parser object

    argv : list
        Command line arguments
    """
    help_cmd = '--help' in argv or '-h' in argv

    # if the file does not exist but the value has sufix yaml/yml, show a
    # warning because the last thing to try is to interpret it as a dotted
    # path and that's probably not what the user wants
    if not entry_point.exists() and entry_point.suffix in {'.yaml', '.yml'}:
        warnings.warn('Entry point value "{}" has extension "{}", which '
                      'suggests a spec file, but the file doesn\'t '
                      'exist'.format(entry_point, entry_point.suffix))

    # even if the entry file is not a file nor a valid module, show the
    # help menu, but show a warning
    if (help_cmd and not entry_point.exists()):
        warnings.warn('Failed to load entry point "{}". It is not a file '
                      'nor a valid dotted path'.format(entry_point))

        args = parser.parse_args()

    # at this point there are two remaining cases:
    # no help command (entry point may or may not exist),:
    #   we attempt to run the command
    # help command and exists:
    #   we just parse parameters to display them in the help menu
    elif entry_point.type == EntryPoint.DottedPath:
        dag, args = parser.process_factory_dotted_path(entry_point)
    elif entry_point.type == EntryPoint.ModulePath:
        dag, args = _process_file_dir_or_glob(
            parser, dagspec_arg=_path_for_module_path(entry_point.value))
    else:
        # process file, directory or glob pattern
        dag, args = _process_file_dir_or_glob(parser)

    return dag, args


def _flatten_dict(d, prefix=''):
    """
    Convert a nested dict: {'a': {'b': 1}} -> {'a--b': 1}
    """
    out = {}

    for k, v in d.items():
        if isinstance(v, Mapping):
            out = {**out, **_flatten_dict(v, prefix=prefix + k + '--')}
        else:
            out[prefix + k] = v

    return out


def _configure_logger(args):
    """Configure logger if user passed --log/--log-file args
    """
    if hasattr(args, 'log'):
        if args.log is not None:
            logging.basicConfig(level=args.log.upper())

    if hasattr(args, "log_file"):
        if args.log_file is not None:
            file_handler = logging.FileHandler(args.log_file)
            logging.getLogger().addHandler(file_handler)



from ploomber.exceptions import ValidationError
from ploomber.io import pretty_print


def keys(valid, passed, required=None, name='spec'):
    passed = set(passed)

    if valid:
        extra = passed - set(valid)

        if extra:
            raise ValidationError(
                "Error validating {}, the following keys aren't "
                "valid: {}. Valid keys are: {}".format(
                    name, pretty_print.iterable(extra),
                    pretty_print.iterable(valid)))

    if required:
        missing = set(required) - passed

        if missing:
            raise ValidationError(f"Error validating {name}. Missing "
                                  f"keys: { pretty_print.iterable(missing)}")



import pydoc
from pathlib import Path, PurePosixPath
from ploomber.placeholders.placeholder import Placeholder
from ploomber.placeholders import extensions
from ploomber.util.util import isiterable_not_str

from jinja2 import Environment, FileSystemLoader, StrictUndefined, exceptions


def _is_iterable_w_types(o, types):
    for element in o:
        if not isinstance(o, types):
            return False
    return True


class SourceLoader:
    """Load source files using a jinja2.Environment

    Data pipelines usually rely on non-Python source code such as SQL scripts,
    SourceLoader provides a convenient way of loading them. This serves two
    purposes: 1) Avoid hardcoded paths to files and 2) Allows using advanced
    jinja2 features that require an Environment object such as macros (under
    the hood, SourceLoader initializes an Environment with a FileSystemLoader)
    SourceLoader returns ploomber.Placeholder objects that can be directly
    passed to Tasks in the source parameter

    Parameters
    ----------
    path : str, pathlib.Path, optional
        Path (or list of) to load files from. Required if module is None

    module : str or module, optional
        Module name as dotted string or module object. Preprends to path
        parameter

    Examples
    --------
    >>> from ploomber import SourceLoader
    >>> loader = SourceLoader('path/to/templates/')
    >>> loader['load_customers.sql'] # doctest: +SKIP
    >>> loader.get_template('load_customers.sql') # doctest: +SKIP
    """
    def __init__(self, path=None, module=None):

        if path is None and module is None:
            raise TypeError('Path cannot be None if module is None')

        # validate path
        if isiterable_not_str(path):
            if _is_iterable_w_types(path, (str, Path)):
                types_found = set(type(element) for element in path)
                raise TypeError('If passing an iterable, path must consist '
                                'of str and pathlib.Path objects only, got '
                                '{}'.format(types_found))

            path = [str(element) for element in path]
        elif isinstance(path, Path):
            path = str(path)

        if isinstance(module, str):
            module_obj = pydoc.locate(module)

            if module_obj is None:
                raise ValueError('Could not locate module "{}"'.format(module))

            module = module_obj

        # for module objects
        if hasattr(module, '__file__'):
            module_path = str(Path(module.__file__).parent.resolve())

        elif module is None:
            module_path = ''
        else:
            raise ValueError('Could not find module path, pass a string or a '
                             'module object')

        if isiterable_not_str(path):
            self.path_full = [str(Path(module_path, e)) for e in path]
        else:
            # if path is None, do not append anything
            self.path_full = str(Path(module_path, path or ''))

        self.env = self.__init_env()

    def __init_env(self):
        # NOTE: we do not use jinja2.PackageLoader since it does not provide
        # the abilty to pass a list of paths
        return Environment(
            loader=FileSystemLoader(self.path_full),
            # this will cause jinja2 to raise an exception if a variable
            # declared in the template is not passed in the render parameters
            undefined=StrictUndefined,
            extensions=(extensions.RaiseExtension, ))

    def __getitem__(self, key):
        return self.get_template(key)

    def get(self, key):
        """Load template, returns None if it doesn' exist
        """
        try:
            return self[key]
        except exceptions.TemplateNotFound:
            return None

    def path_to(self, key):
        """Return the path to a template, even if it doesn't exist
        """
        try:
            return self[key].path
        except exceptions.TemplateNotFound:
            return Path(self.path_full, key)

    def get_template(self, name):
        """Load a template by name

        Parameters
        ----------
        name : str or pathlib.Path
            Template to load
        """
        # if name is a nested path, this will return an appropriate
        # a/b/c string even on Windows
        path = str(PurePosixPath(*Path(name).parts))

        try:
            template = self.env.get_template(path)
        except exceptions.TemplateNotFound as e:
            exception = e
        else:
            exception = None

        if exception is not None:
            expected_path = str(Path(self.path_full, name))

            # user saved the template locally, but the source loader is
            # configured to load from a different place
            if Path(name).exists():
                raise exceptions.TemplateNotFound(
                    f'{str(name)!r} template does not exist. '
                    'However such a file exists in the current working '
                    'directory, if you want to load it as a template, move it '
                    f'to {self.path_full!r} or remove the source_loader')
            # no template and the file does not exist, raise a generic message
            else:
                raise exceptions.TemplateNotFound(
                    f'{str(name)!r} template does not exist. '
                    'Based on your configuration, if should be located '
                    f'at: {expected_path!r}')

        return Placeholder(template)

    def _ipython_key_completions_(self):
        return self.env.list_templates()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['env']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.env = self.__init_env()



"""
Path management
---------------
Relative paths are heavily used for pointing to source files and products,
when only the Python API existed, this was unambiguous: all paths
were relative to the current working directory, you only had to make sure
to start the session at the right place. Initially, the Spec API didn't
change a thing since the assumption was that one would convert a spec to
a DAG in a folder containing a pipeline.yaml, under this scenario the
current directory is also the parent of pipeline.yaml so all relative paths
work as expected.

However, when we introduced the Jupyter custom contents manager things
changed because the DAG can be initialized from directories where there
isn't a pipeline.yaml but we have to recursively look for one. Furthermore,
if extract_upstream or extract_product are True and the source files do
not share the parent with pipeline.yaml an ambiguity arises: are paths
inside a source file relative to the source file, current directory or
pipeline.yaml?

When we started working on Soopervisor, the concept of "project" arised,
which is a folder that contains all necessary files to instantiate a
Ploomber pipeline. To take current directory ambiguity out of the situation
we decided that projects that use the Spec API should ignore the current
directory and assume that *all relative paths in the spec are so to the
pipeline.yaml parent folder* this means that we could initialize projects
from anywhere by just pointing to the project's root folder (the one that
contains pipeline.yaml). This is implemented in TaskSpec.py@init_product
where we add the project's root folder to all relative paths. This implies
that all relative paths are converted to absolute to make them agnostic
from the current working directory. The basic idea is that a person
writing a DAG spec should only have to know that paths are relative
to the project's root folder and the DAGSpec should do the heavy lifting
of instantiating a spec that could be converted to a dag (and reendered)
independent of the working directory, because the project's root folder
defines the pipelines's scope, and there is no point in looking for files
outside of it, except when asking explicity through an absolute path.

However, this behavior clashes with the Jupyter notebook defaults. Jupyter
sets the current working directory not to the directory where
"jupyter {notebook/lab}" was called but to the current file parent's.
Our current logic would override this default behavior so we added an
option to turn it back on (product_relative_to_source).

Note that this does not affect the Python API, since that would create more
confusion. The Python API should work as expected from any other library:
paths should be relative to the current working directory.

The only case where we can't determine a project's root is when a DAGSpec
is initialized from a dictionary directly, but this is not something
that end-users would do.

Spec sections
-------------
"meta" is for settings that cannot be clearly mapped to the OOP Python
interface, we don't call it config because there is a DAGConfig object in
the Python API and this might cause confusion

All other sections should represent valid DAG properties.


Partials [Provisional name]
---------------------------
Partials were introduced to support a common ML use case: train and serving
pipelines. If we describe a training pipeline as a DAG, the only difference
with its serving counterpart are the tasks (can be more than one) at the
beginning (train: get historical data, serve: get single data point) and the
end (train: fit a model, serve: load a model and predict). Everything that
happens in the middle should be the same to avoid training-serving skew. There
are two specific use cases: batch and online.

To define a batch serving pipeline we define three files: pipeline.yaml
(train DAG), pipeline-serve (serve DAG), and pipeline-features.yaml (partial).
The first two follow the DAGSpec schema and use
``import_tasks_from: pipeline-features.yaml``. Training is done via
``ploomber build --entry-point pipeline.yaml`` and serving via
``ploomber build --entry-point pipeline-serve.yaml``.

An online API has a different mechanism because the inference pipeline needs
a Python API to interface with a web framework, rpc, or similar. This logic
is implemented by OnlineDAG, which takes a partial definition
(e.g. ``pipeline-features.yaml``) and builds an in-memory DAG that can make
predictions using ``OnlineDAG().predict()``. See ``OnlineDAG`` documentation
for details.
"""
import click
import fnmatch
import os
import yaml
import logging
from pathlib import Path
from collections.abc import MutableMapping, Mapping
from glob import iglob
from itertools import chain
import pprint
import warnings

from ploomber.dag.dag import DAG
from ploomber.placeholders.sourceloader import SourceLoader
from ploomber.util.util import call_with_dictionary, add_to_sys_path
from ploomber.util import dotted_path
from ploomber.spec.taskspec import TaskSpec, suffix2taskclass
from ploomber.util import validate
from ploomber.util import default
from ploomber.dag.dagconfiguration import DAGConfiguration
from ploomber.exceptions import (DAGSpecInitializationError,
                                 MissingParametersCellError)
from ploomber.env.envdict import EnvDict
from ploomber.env.expand import (expand_raw_dictionary_and_extract_tags,
                                 expand_raw_dictionaries_and_extract_tags)
from ploomber.tasks import NotebookRunner
from ploomber.tasks.taskgroup import TaskGroup
from ploomber.validators.string import (validate_product_class_name,
                                        validate_task_class_name)
from ploomber.executors import Parallel
from ploomber.io import pretty_print
from ploomber.sources import notebooksource

logger = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=4)


class DAGSpec(MutableMapping):
    """
    A DAG spec is a dictionary with certain structure that can be converted
    to a DAG using DAGSpec.to_dag().

    There are two cases: the simplest one is just a dictionary with a
    "location" key with the factory to call, the other explicitly describes
    the DAG structure as a dictionary.

    When .to_dag() is called, the current working directory is temporarily
    switched to the spec file parent folder (only applies when loading from
    a file)

    Parameters
    ----------
    data : str, pathlib.Path or dict
        Path to a YAML spec or dict spec. If loading from a file, sources
        and products are resolved to the file's parent. If the file is in a
        packaged structure (i.e., src/package/pipeline.yaml), the existence
        of a setup.py in the same folder as src/ is validated. If loaded from
        a dict, sources and products aren't resolved, unless a parent_path
        is not None.

    env : dict, pathlib.path or str, optional
        If path it must be a YAML file. Environment to load. Any string with
        the format '{{placeholder}}' in the spec is replaced by the
        corresponding value in the given key (i.e., placeholder). If ``env``
        is None and spec is a dict, no env is loaded. If None and loaded
        from a YAML file, an ``env.yaml`` file is loaded from the current
        working diirectory, if it doesn't exist, it is loaded from the YAML
        spec parent folder. If none of these exist, no env is loaded.
        A :py:mod:`ploomber.Env` object is initialized, see documentation for
        details.

    lazy_import : bool, optional
        Whether to import dotted paths to initialize PythonCallables with the
        actual function. If False, PythonCallables are initialized directly
        with the dotted path, which means some verifications such as import
        statements in that function's module are delayed until the pipeline
        is executed. This also applies to placeholders loaded using a
        SourceLoader, if a template exists, it will return the path
        to it, instead of initializing it, if it doesn't, it will return None
        instead of raising an error. This setting is useful when we require to
        load YAML spec and instantiate the DAG object to extract information
        from it (e.g., which are the declared tasks) but the process running
        it may not have all the required dependencies to do so (e.g., an
        imported library in a PythonCallable task).

    reload : bool, optional
        Reloads modules before importing dotted paths to detect code changes
        if the module has already been imported. Has no effect if
        lazy_import=True.

    Attributes
    ----------
    path : str or None
        Returns the path used to load the data. None if loaded from a
        dictionary
    """

    # NOTE: lazy_import is used where we need to initialized a a spec but don't
    # plan on running it. One use case is when exporting to Argo or Airflow:
    # we don't want to raise errors if some dependency is missing because
    # it can happen that the environment exporting the dag does not have
    # all the dependencies required to run it. The second use case is when
    # running "ploomber scaffold", we want to use DAGSpec machinery to parse
    # the yaml spec and the use such information to add the task in the
    # appropriate place, this by construction, means that the spec as it is
    # cannot be converted to a dag yet, since it has at least one task
    # whose source does not exist
    def __init__(self,
                 data,
                 env=None,
                 lazy_import=False,
                 reload=False,
                 parent_path=None):
        self._init(data=data,
                   env=env,
                   lazy_import=lazy_import,
                   reload=reload,
                   parent_path=parent_path,
                   look_up_project_root_recursively=True)

    def _init(self, data, env, lazy_import, reload, parent_path,
              look_up_project_root_recursively):
        self._lazy_import = lazy_import
        self._name = None

        # initialized with a path to a yaml file...
        if isinstance(data, (str, Path)):
            # TODO: test this
            if parent_path is not None:
                raise ValueError('parent_path must be None when '
                                 f'initializing {type(self).__name__} with '
                                 'a path to a YAML spec')
            # resolve the parent path to make sources and products unambiguous
            # even if the current working directory changes
            self._path = Path(data).resolve()
            self._parent_path = str(self._path.parent)

            # assign the name of the parent directory
            self._name = self._path.parent.name

            if not Path(data).is_file():
                raise FileNotFoundError(
                    'Error initializing DAGSpec with argument '
                    f'{data!r}: Expected it to be a path to a YAML file, but '
                    'such file does not exist')

            content = Path(data).read_text()

            try:
                data = yaml.safe_load(content)
            except (
                    yaml.parser.ParserError,
                    yaml.constructor.ConstructorError,
                    yaml.scanner.ScannerError,
            ) as e:
                error = e
            else:
                error = None

            if error:
                if '{{' in content or '}}' in content:
                    raise DAGSpecInitializationError(
                        'Failed to initialize spec. It looks like '
                        'you\'re using placeholders (i.e. {{placeholder}}). '
                        'Make sure values are enclosed in parentheses '
                        '(e.g. key: "{{placeholder}}"). Original '
                        'parser error:\n\n'
                        f'{error}')
                else:
                    raise DAGSpecInitializationError(
                        'Failed to initialize spec. Got invalid YAML'
                    ) from error

        # initialized with a dictionary...
        else:
            self._path = None
            # FIXME: add test cases, some of those features wont work if
            # _parent_path is None. We should make sure that we either raise
            # an error if _parent_path is needed or use the current working
            # directory if it's appropriate - this is mostly to make relative
            # paths consistent: they should be relative to the file that
            # contains them
            self._parent_path = (None if not parent_path else str(
                Path(parent_path).resolve()))

        self.data = data

        if isinstance(self.data, list):
            self.data = {'tasks': self.data}

        if self.data.get('tasks') and not isinstance(self.data['tasks'], list):
            raise DAGSpecInitializationError(
                'Expected \'tasks\' in the dag spec to contain a '
                f'list, but got: {self.data["tasks"]} '
                '(an object with '
                f'type: {type(self.data["tasks"]).__name__!r})')

        # validate keys defined at the top (nested keys are not validated here)
        self._validate_top_keys(self.data, self._path)

        logger.debug('DAGSpec enviroment:\n%s', pp.pformat(env))

        env = env or dict()
        path_to_defaults = default.path_to_env_from_spec(
            path_to_spec=self._path)

        # there is an env.yaml we can use
        if path_to_defaults:
            defaults = yaml.safe_load(Path(path_to_defaults).read_text())
            self.env = EnvDict(env,
                               path_to_here=self._parent_path,
                               defaults=defaults)

        # there is no env.yaml
        else:
            self.env = EnvDict(env, path_to_here=self._parent_path)

        self.data, tags = expand_raw_dictionary_and_extract_tags(
            self.data, self.env)

        logger.debug('Expanded DAGSpec:\n%s', pp.pformat(data))

        # if there is a "location" top key, we don't have to do anything else
        # as we will just load the dotted path when .to_dag() is called
        if 'location' not in self.data:

            Meta.initialize_inplace(self.data)

            import_tasks_from = self.data['meta']['import_tasks_from']

            if import_tasks_from is not None:
                # when using a relative path in "import_tasks_from", we must
                # make it absolute...
                if not Path(import_tasks_from).is_absolute():
                    # use _parent_path if there is one
                    if self._parent_path:
                        self.data['meta']['import_tasks_from'] = str(
                            Path(self._parent_path, import_tasks_from))
                    # otherwise just make it absolute
                    else:
                        self.data['meta']['import_tasks_from'] = str(
                            Path(import_tasks_from).resolve())

                imported = yaml.safe_load(
                    Path(self.data['meta']['import_tasks_from']).read_text())

                if not imported:
                    path = str(self.data['meta']['import_tasks_from'])
                    raise ValueError('expected import_tasks_from file '
                                     f'({path!r}) to return a list of tasks, '
                                     f'got: {imported}')

                if not isinstance(imported, list):
                    raise TypeError(
                        'Expected list when loading YAML file from '
                        'import_tasks_from: file.yaml, '
                        f'but got {type(imported)}')

                if self.env is not None:
                    (imported,
                     tags_other) = expand_raw_dictionaries_and_extract_tags(
                         imported, self.env)
                    tags = tags | tags_other

                # relative paths here are relative to the file where they
                # are declared
                base_path = Path(self.data['meta']['import_tasks_from']).parent

                for task in imported:
                    add_base_path_to_source_if_relative(task,
                                                        base_path=base_path)

                self.data['tasks'].extend(imported)

            # check if there are any params declared in env, not used in
            # in the pipeline
            extra = set(self.env) - self.env.default_keys - tags

            if extra:
                warnings.warn('The following placeholders are declared in the '
                              'environment but '
                              f'unused in the spec: {extra}')

            if self.data['tasks'] is None:
                raise DAGSpecInitializationError(
                    'Failed to initialize spec, "tasks" section is empty')

            self.data['tasks'] = [
                normalize_task(task) for task in self.data['tasks']
            ]

            # NOTE: for simple projects, project root is the parent folder
            # of pipeline.yaml, for package projects is the parent folder
            # of setup.py
            if look_up_project_root_recursively:
                project_root = (
                    None if not self._parent_path else
                    default.find_root_recursively(
                        starting_dir=self._parent_path,
                        filename=None if not self._path else self._path.name))
            else:
                project_root = self._parent_path

            # make sure the folder where the pipeline is located is in sys.path
            # otherwise dynamic imports needed by TaskSpec will fail
            with add_to_sys_path(self._parent_path, chdir=False):
                self.data['tasks'] = [
                    TaskSpec(t,
                             self.data['meta'],
                             project_root=project_root,
                             lazy_import=lazy_import,
                             reload=reload) for t in self.data['tasks']
                ]
        else:
            self.data['meta'] = Meta.empty()

    def _validate_top_keys(self, spec, path):
        """Validate keys at the top of the spec
        """
        if 'tasks' not in spec and 'location' not in spec:
            raise DAGSpecInitializationError(
                'Failed to initialize spec. Missing "tasks" key')

        if 'location' in spec:
            if len(spec) > 1:
                raise DAGSpecInitializationError(
                    'Failed to initialize spec. If '
                    'using the "location" key there should not '
                    'be other keys')

        else:
            valid = {
                'meta',
                'config',
                'clients',
                'tasks',
                'serializer',
                'unserializer',
                'executor',
                'on_finish',
                'on_render',
                'on_failure',
            }
            validate.keys(valid, spec.keys(), name='dag spec')

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        for key in self.data:
            yield key

    def __len__(self):
        return len(self.data)

    def to_dag(self):
        """Converts the DAG spec to a DAG object
        """
        # when initializing DAGs from pipeline.yaml files, we have to ensure
        # that the folder where pipeline.yaml is located is in sys.path for
        # imports to work (for dag clients), this happens most of the time but
        # for some (unknown) reason, it doesn't
        # happen when initializing PloomberContentsManager.
        # pipeline.yaml paths are written relative to that file, for source
        # scripts to be located we temporarily change the current working
        # directory
        with add_to_sys_path(self._parent_path, chdir=True):
            dag = self._to_dag()

        return dag

    def _to_dag(self):
        """
        Internal method to manage the different cases to convert to a DAG
        object
        """
        if 'location' in self:
            return dotted_path.call_dotted_path(self['location'])

        dag = DAG(name=self._name)

        if 'config' in self:
            dag._params = DAGConfiguration.from_dict(self['config'])

        if 'executor' in self:
            executor = self['executor']

            if isinstance(executor,
                          str) and executor in {'serial', 'parallel'}:
                if executor == 'parallel':
                    dag.executor = Parallel()
            elif isinstance(executor, Mapping):
                dag.executor = dotted_path.DottedPath(
                    executor, lazy_load=False, allow_return_none=False)()
            else:
                raise DAGSpecInitializationError(
                    '"executor" must be '
                    '"serial", "parallel", or a dotted path'
                    f', got: {executor!r}')

        clients = self.get('clients')

        if clients:
            for class_name, dotted_path_spec in clients.items():
                dps = dotted_path.DottedPath(dotted_path_spec,
                                             lazy_load=self._lazy_import,
                                             allow_return_none=False)

                if self._lazy_import:
                    dag.clients[class_name] = dps
                else:
                    dag.clients[class_name] = dps()

        for attr in ('serializer', 'unserializer', 'on_finish', 'on_render',
                     'on_failure'):
            if attr in self:
                setattr(
                    dag, attr,
                    dotted_path.DottedPath(self[attr],
                                           lazy_load=self._lazy_import))

        process_tasks(dag, self, root_path=self._parent_path)

        return dag

    @property
    def path(self):
        return self._path

    @classmethod
    def _find_relative(cls, name=None, lazy_import=False):
        """
        Searches for a spec in default locations relative to the current
        working directory

        Notes
        -----
        This is a private API used by Soopervisor to locate which YAML spec
        to use for exporting. It needs a relative path since the location
        is used for loading the DAG and submitting tasks and as an argument
        in "ploomber task --entry-point {relative-path}" when executing tasks
        """
        relative_path = default.entry_point_relative(name=name)
        return cls(relative_path, lazy_import=lazy_import), relative_path

    @classmethod
    def find(cls,
             env=None,
             reload=False,
             lazy_import=False,
             starting_dir=None,
             name=None):
        """
        Automatically find pipeline.yaml and return a DAGSpec object, which
        can be converted to a DAG using .to_dag()

        Parameters
        ----------
        env
            The environment to pass to the spec

        name : str, default=None
            Filename to search for. If None, it looks for a pipeline.yaml file,
            otherwise it looks for a file with such name.
        """
        starting_dir = starting_dir or os.getcwd()
        path_to_entry_point = default.entry_point_with_name(
            root_path=starting_dir, name=name)

        try:
            return cls(path_to_entry_point,
                       env=env,
                       lazy_import=lazy_import,
                       reload=reload)
        except Exception as e:
            exc = DAGSpecInitializationError('Error initializing DAG from '
                                             f'{path_to_entry_point!s}')
            raise exc from e

    @classmethod
    def from_directory(cls, path_to_dir):
        """
        Construct a DAGSpec from a directory. Product and upstream are
        extracted from sources

        Parameters
        ----------
        path_to_dir : str
            The directory to use.  Looks for scripts
            (``.py``, ``.R`` or ``.ipynb``) in the directory and interprets
            them as task sources, file names are assigned as task names
            (without extension). The spec is generated with the default values
            in the "meta" section. Ignores files with invalid extensions.

        Notes
        -----
        ``env`` is not supported because the spec is generated from files
        in ``path_to_dir``, hence, there is no way to embed tags
        """
        valid_extensions = [
            name for name, class_ in suffix2taskclass.items()
            if class_ is NotebookRunner
        ]

        if Path(path_to_dir).is_dir():
            pattern = str(Path(path_to_dir, '*'))
            files = list(
                chain.from_iterable(
                    iglob(pattern + ext) for ext in valid_extensions))
            return cls.from_files(files)
        else:
            raise NotADirectoryError(f'{path_to_dir!r} is not a directory')

    @classmethod
    def from_files(cls, files):
        """
        Construct DAGSpec from list of files or glob-like pattern. Product and
        upstream are extracted from sources

        Parameters
        ----------
        files : list or str
            List of files to use or glob-like string pattern. If glob-like
            pattern, ignores directories that match the criteria.
        """
        valid_extensions = [
            name for name, class_ in suffix2taskclass.items()
            if class_ is NotebookRunner
        ]

        if isinstance(files, str):
            files = [f for f in iglob(files) if Path(f).is_file()]

        invalid = [f for f in files if Path(f).suffix not in valid_extensions]

        if invalid:
            raise ValueError(f'Cannot instantiate DAGSpec from files with '
                             f'invalid extensions: {invalid}. '
                             f'Allowed extensions are: {valid_extensions}')

        tasks = [{'source': file_} for file_ in files]
        meta = {'extract_product': True, 'extract_upstream': True}
        return cls({'tasks': tasks, 'meta': meta})


class DAGSpecPartial(DAGSpec):
    """
    A DAGSpec subclass that initializes from a list of tasks (used in the
    onlinedag.py) module
    """
    def __init__(self, path_to_partial, env=None):
        with open(path_to_partial) as f:
            tasks = yaml.safe_load(f)

        if not isinstance(tasks, list):
            raise ValueError(
                f'Expected partial {path_to_partial!r} to be a '
                f'list of tasks, but got a {type(tasks).__name__} instead')

        # cannot extract upstream because this is an incomplete DAG
        meta = {'extract_product': False, 'extract_upstream': False}
        data = {'tasks': tasks, 'meta': meta}

        env = env or default.path_to_env_from_spec(path_to_partial)

        self._init(data=data,
                   env=env,
                   lazy_import=False,
                   reload=False,
                   parent_path=Path(path_to_partial).parent,
                   look_up_project_root_recursively=False)


class Meta:
    """Schema for meta section in pipeline.yaml
    """
    VALID = {
        'extract_upstream',
        'extract_product',
        'product_default_class',
        'product_relative_to_source',
        'jupyter_hot_reload',
        'source_loader',
        'jupyter_functions_as_notebooks',
        'import_tasks_from',
    }

    @classmethod
    def initialize_inplace(cls, data):
        """Validate and instantiate the "meta" section
        """
        if 'meta' not in data:
            data['meta'] = {}

        data['meta'] = Meta.default_meta(data['meta'])

    @classmethod
    def default_meta(cls, meta=None):
        """Fill missing values in a meta dictionary
        """
        if meta is None:
            meta = {}

        validate.keys(cls.VALID, meta, name='dag spec')

        if 'extract_upstream' not in meta:
            meta['extract_upstream'] = True

        if 'extract_product' not in meta:
            meta['extract_product'] = False

        if 'product_relative_to_source' not in meta:
            meta['product_relative_to_source'] = False

        if 'jupyter_hot_reload' not in meta:
            meta['jupyter_hot_reload'] = True

        if 'jupyter_functions_as_notebooks' not in meta:
            meta['jupyter_functions_as_notebooks'] = False

        if 'import_tasks_from' not in meta:
            meta['import_tasks_from'] = None

        if 'source_loader' not in meta:
            meta['source_loader'] = None
        else:
            try:
                meta['source_loader'] = SourceLoader(**meta['source_loader'])
            except Exception as e:
                msg = ('Error initializing SourceLoader with '
                       f'{meta["source_loader"]}. Error message: {e.args[0]}')
                e.args = (msg, )
                raise

        defaults = {
            'SQLDump': 'File',
            'NotebookRunner': 'File',
            'ScriptRunner': 'File',
            'SQLScript': 'SQLRelation',
            'PythonCallable': 'File',
            'ShellScript': 'File',
        }

        if 'product_default_class' not in meta:
            meta['product_default_class'] = defaults
        else:
            for class_, prod in defaults.items():
                if class_ not in meta['product_default_class']:
                    meta['product_default_class'][class_] = prod

        # validate keys and values in product_default_class
        for task_name, product_name in meta['product_default_class'].items():
            try:
                validate_task_class_name(task_name)
                validate_product_class_name(product_name)
            except Exception as e:
                msg = f'Error validating product_default_class: {e.args[0]}'
                e.args = (msg, )
                raise

        return meta

    @classmethod
    def empty(cls):
        """Default meta values when a spec is initialized with "location"
        """
        return {v: None for v in cls.VALID}


def process_tasks(dag, dag_spec, root_path=None):
    """
    Initialize Task objects from TaskSpec, extract product and dependencies
    if needed and set the dag dependencies structure
    """
    root_path = root_path or '.'

    # options
    extract_up = dag_spec['meta']['extract_upstream']
    extract_prod = dag_spec['meta']['extract_product']

    # raw values extracted from the upstream key
    upstream_raw = {}

    # first pass: init tasks and them to dag
    for task_dict in dag_spec['tasks']:
        # init source to extract product
        fn = task_dict['class']._init_source
        kwargs = {'kwargs': {}, **task_dict}

        try:
            source = call_with_dictionary(fn, kwargs=kwargs)
        except MissingParametersCellError:
            missing_params = True
        else:
            missing_params = False

        if missing_params:
            click.secho(
                f'{kwargs["source"]} is missing the parameters cell, '
                'adding it at the top of the file...',
                fg='yellow')
            notebooksource.add_parameters_cell(kwargs['source'], extract_up,
                                               extract_prod)
            source = call_with_dictionary(fn, kwargs=kwargs)

        if extract_prod:
            task_dict['product'] = source.extract_product()

        # convert to task, up has the content of "upstream" if any
        task, up = task_dict.to_task(dag)

        if isinstance(task, TaskGroup):
            for t in task:
                upstream_raw[t] = up
        else:
            if extract_prod:
                logger.debug('Extracted product for task "%s": %s', task.name,
                             task.product)
            upstream_raw[task] = up

    # second optional pass: extract upstream
    tasks = list(dag.values())
    task_names = list(dag._iter())
    # actual upstream values after matching wildcards
    upstream = {}

    # expand upstream dependencies (in case there are any wildcards)
    for task in tasks:
        if extract_up:
            try:
                extracted = task.source.extract_upstream()
            except Exception as e:
                raise DAGSpecInitializationError(
                    f'Failed to initialize task {task.name!r}') from e

            upstream[task] = _expand_upstream(extracted, task_names)
        else:
            upstream[task] = _expand_upstream(upstream_raw[task], task_names)

        logger.debug('Extracted upstream dependencies for task %s: %s',
                     task.name, upstream[task])

    # Last pass: set upstream dependencies
    for task in tasks:
        if upstream[task]:
            for task_name, group_name in upstream[task].items():

                up = dag.get(task_name)

                if up is None:
                    names = [t.name for t in tasks]
                    raise DAGSpecInitializationError(
                        f'Task {task.name!r} '
                        'has an upstream dependency '
                        f'{task_name!r}, but such task '
                        'doesn\'t exist. Available tasks: '
                        f'{pretty_print.iterable(names)}')

                task.set_upstream(up, group_name=group_name)


def normalize_task(task):
    # NOTE: we haven't documented that we support tasks as just strings,
    # should we keep this or deprecate it?
    if isinstance(task, str):
        return {'source': task}
    else:
        return task


def add_base_path_to_source_if_relative(task, base_path):
    path = Path(task['source'])
    relative_source = not path.is_absolute()

    # must be a relative source with a valid extension, otherwise, it can
    # be a dotted path
    if relative_source and path.suffix in set(suffix2taskclass):
        task['source'] = str(Path(base_path, task['source']).resolve())


def _expand_upstream(upstream, task_names):
    """
    Processes a list of upstream values extracted from source (or declared in
    the spec's "upstream" key). Expands wildcards like "some-task-*" to all
    the values that match. Returns a dictionary where keys are the upstream
    dependencies and the corresponding value is the wildcard. If no wildcard,
    the value is None
    """
    if not upstream:
        return None

    expanded = {}

    for up in upstream:
        if '*' in up:
            matches = fnmatch.filter(task_names, up)
            expanded.update({match: up for match in matches})
        else:
            expanded[up] = None

    return expanded



from pathlib import Path
import os

from ploomber.spec import DAGSpec
from ploomber.util import default
from ploomber.util.dotted_path import load_callable_dotted_path
from ploomber.exceptions import DAGSpecInitializationError
from ploomber.entrypoint import try_to_find_entry_point_type, EntryPoint


# NOTE: this function is importable in the top-level package to allow
# users debug errors when using the Jupyter integration
def lazily_load_entry_point(starting_dir=None, reload=False):
    """
    Lazily loads entry point by recursively looking in starting_dir directory
    and parent directories.
    """

    starting_dir = starting_dir or '.'

    entry_point = os.environ.get('ENTRY_POINT')

    type_ = try_to_find_entry_point_type(entry_point)

    if type_ == EntryPoint.Directory:
        spec = DAGSpec.from_directory(entry_point)
        path = Path(entry_point)
    elif type_ == EntryPoint.DottedPath:
        entry = load_callable_dotted_path(str(entry_point), raise_=True)
        dag = entry()
        spec = dict(meta=dict(jupyter_hot_reload=False,
                              jupyter_functions_as_notebooks=False))
        # potential issue: dag defines sources as relative paths
        path = Path().resolve()
        return spec, dag, path
    else:
        spec, path, _ = _default_spec_load(starting_dir=starting_dir,
                                           reload=reload,
                                           lazy_import=True)

    # chain exception to provide more context
    dag = spec.to_dag()

    # we remove the on_render hook because this is a lazy load, if we don't do
    # it, calling the hook will cause an error since the function never loads
    dag.on_render = None

    # same with task-level hooks
    # also disable static_analysis since we don't want to break cell injection
    # because of some issues in te code
    for name in dag._iter():
        task = dag[name]
        task._on_render = None

        if hasattr(task, 'static_analysis'):
            task.static_analysis = False

    return spec, dag, path


def _default_spec_load(starting_dir=None, lazy_import=False, reload=False):
    """
    NOTE: this is a private API. Use DAGSpec.find() instead

    Looks for a pipeline.yaml, generates a DAGSpec and returns a DAG.
    Currently, this is only used by the PloomberContentsManager, this is
    not intended to be a public API since initializing specs from paths
    where we have to recursively look for a pipeline.yaml has some
    considerations regarding relative paths that make this confusing,
    inside the contents manager, all those things are all handled for that
    use case.

    The pipeline.yaml parent folder is temporarily added to sys.path when
    calling DAGSpec.to_dag() to make sure imports work as expected

    Returns DAG and the directory where the pipeline.yaml file is located.
    """
    root_path = starting_dir or os.getcwd()
    path_to_entry_point = default.entry_point(root_path=root_path)

    try:
        spec = DAGSpec(path_to_entry_point,
                       env=None,
                       lazy_import=lazy_import,
                       reload=reload)

        path_to_spec = Path(path_to_entry_point)
        return spec, path_to_spec.parent, path_to_spec

    except Exception as e:
        exc = DAGSpecInitializationError('Error initializing DAG from '
                                         f'{path_to_entry_point!s}')
        raise exc from e



"""
Validation functions
"""
from collections.abc import Mapping


def raw_data_keys(d):
    """
    Validate raw dictionary, no top-level keys with leading underscores,
    except for _module, and no keys with double underscores anywhere
    """
    keys_all = get_keys_for_dict(d)
    errors = []

    try:
        no_double_underscores(keys_all)
    except ValueError as e:
        errors.append(str(e))

    try:
        no_leading_underscore(d.keys())
    except ValueError as e:
        errors.append(str(e))

    if errors:
        msg = 'Error validating env.\n' + '\n'.join(errors)
        raise ValueError(msg)


def get_keys_for_dict(d):
    """
    Get all (possibly nested) keys in a dictionary
    """
    out = []

    for k, v in d.items():
        out.append(k)

        if isinstance(v, Mapping):
            out += get_keys_for_dict(v)

    return out


def no_double_underscores(keys):
    not_allowed = [k for k in keys if '__' in k]

    if not_allowed:
        raise ValueError('Keys cannot have double '
                         'underscores, got: {}'.format(not_allowed))


def no_leading_underscore(keys):
    allowed = {'_module'}
    not_allowed = [k for k in keys if k.startswith('_') and k
                   not in allowed]

    if not_allowed:
        raise ValueError('Top-level keys cannot start '
                         'with an underscore, except for {}. '
                         'Got: {}'.format(allowed, not_allowed))



import warnings
import abc
from collections.abc import Mapping

import yaml


class Config(abc.ABC):
    """An abstract class to create configuration files (stored as YAML)

    Notes
    -----
    For examples, see test_config.py or the concrete classes
    (UserSettings, Internal)
    """
    def __init__(self):
        self._init_values()

        # resolve home directory
        path = self.path()

        if not path.exists():
            defaults = self._get_data()
            path.write_text(yaml.dump(defaults))
            self._set_data(defaults)
        else:
            try:
                content = self._load_from_file()
                loaded = True
            except Exception as e:
                warnings.warn(f'Error loading {str(path)!r}: {e}\n\n'
                              'reverting to default values')
                loaded = False
                content = self._get_data()

            if loaded and not isinstance(content, Mapping):
                warnings.warn(
                    f'Error loading {str(path)!r}. Expected a dictionary '
                    f'but got {type(content).__name__}, '
                    'reverting to default values')
                content = self._get_data()

            self._set_data(content)

    def _load_from_file(self):
        path = self.path()
        text = path.read_text()
        content = yaml.safe_load(text)

        for key, type_ in self.__annotations__.items():
            value = content.get(key, None)

            if value is not None and not isinstance(value, type_):
                default = getattr(self, key)
                warnings.warn(f'Corrupted config file {str(path)!r}: '
                              f'expected {key!r} to contain an object '
                              f'with type {type_.__name__}, but got '
                              f'{type(value).__name__}. Reverting to '
                              f'default value {default}')
                content[key] = default

        return content

    def _get_data(self):
        """Extract values from the annotations and return a dictionary
        """
        return {key: getattr(self, key) for key in self.__annotations__}

    def _set_data(self, data):
        """Take a dictionary and store it in the annotations
        """
        for key in self.__annotations__:
            if key in data:
                setattr(self, key, data[key])

    def _init_values(self):
        """
        Iterate over annotations to initialize values. This is only relevant
        when any of the annotations has a factory method to initialize the
        values. If they value is a literal, no changes happen.
        """
        for key in self.__annotations__:
            name = f'{key}_default'

            # if there is a method with such name, call it and store the output
            if hasattr(self, name):
                value = getattr(self, name)()
                # call __setattr__ on the superclass so we skip the part
                # where we overwrite the YAML file, here we only want to
                # set the default values
                super().__setattr__(key, value)

    def _write(self):
        """Writes data to the YAML file
        """
        data = self._get_data()
        self.path().write_text(yaml.dump(data))

    def __setattr__(self, name, value):
        if name not in self.__annotations__:
            raise ValueError(f'{name} not a valid field')
        else:
            super().__setattr__(name, value)
            self._write()

    @abc.abstractclassmethod
    def path(cls):
        """Returns the path to the YAML file
        """
        pass



from ploomber.tasks.tasks import (PythonCallable, ShellScript, DownloadFromURL,
                                  Link, Input, task_factory)
from ploomber.tasks.taskfactory import TaskFactory
from ploomber.tasks.sql import (SQLScript, SQLDump, SQLTransfer, SQLUpload,
                                PostgresCopyFrom)
from ploomber.tasks.notebook import NotebookRunner, ScriptRunner
from ploomber.tasks.aws import UploadToS3
from ploomber.tasks.param_forward import input_data_passer, in_memory_callable
from ploomber.tasks.taskgroup import TaskGroup
from ploomber.tasks.abc import Task

__all__ = [
    'Task',
    'PythonCallable',
    'ShellScript',
    'TaskFactory',
    'SQLScript',
    'SQLDump',
    'SQLTransfer',
    'SQLUpload',
    'PostgresCopyFrom',
    'NotebookRunner',
    'ScriptRunner',
    'DownloadFromURL',
    'Link',
    'Input',
    'UploadToS3',
    'TaskGroup',
    'input_data_passer',
    'in_memory_callable',
    'task_factory',
]



import sys


def debug_if_exception(callable_):
    """
    Drop a debugger session if running callable_() raises an exception,
    otherwise it just returns the value returned by callable_()
    """
    # NOTE: importing it here, otherwise we get a
    # "If you suspect this is an IPython X.Y.Z bug..." message if any exception
    # after the import if an exception happens
    # NOTE: the IPython.terminal.debugger module has pdb-like classes but it
    # doesn't mimic pdb's API exactly, ipdb is just a wrapper that takes care
    # of those details - I tried using IPython directly but bumped into some
    # issues
    import ipdb

    try:
        result = callable_()
    except Exception:
        ipdb.post_mortem(sys.exc_info()[2])
    else:
        return result



import sys


def _python_bin():
    """
    Get the path to the Python executable, return 'python' if unable to get it
    """
    executable = sys.executable
    return executable if executable else 'python'



import ast
from pathlib import Path

import jupytext
from jinja2 import Environment, PackageLoader, StrictUndefined

from ploomber import tasks
from ploomber.util.dotted_path import (locate_dotted_path,
                                       create_intermediate_modules)


class ScaffoldLoader:
    """Scaffold task files
    """
    def __init__(self):
        self.env = Environment(loader=PackageLoader(
            'ploomber', str(Path('resources', 'ploomber_add'))),
                               variable_start_string='[[',
                               variable_end_string=']]',
                               block_start_string='[%',
                               block_end_string='%]',
                               undefined=StrictUndefined)

    def get_template(self, name):
        return self.env.get_template(name)

    def render(self, name, params):
        # if requested the ipynb template, we must convert it from the .py file
        if name == 'task.ipynb':
            p = Path(name)
            convert_to = p.suffix[1:]
            name = str(p.with_suffix('.py'))
            is_ipynb = True
        else:
            convert_to = None
            is_ipynb = False

        t = self.get_template(name)
        out = t.render(**params, is_ipynb=is_ipynb)

        if convert_to:
            nb = jupytext.reads(out, fmt='py:percent')
            out = jupytext.writes(nb, fmt=convert_to)

        return out

    def create(self, source, params, class_):
        """Scaffold a task if they don't exist

        Returns
        -------
        bool
            True if it created a task, False if it didn't
        """
        did_create = False

        if class_ is tasks.PythonCallable:
            source_parts = source.split('.')
            (*module_parts, fn_name) = source_parts
            params['function_name'] = fn_name

            try:
                spec = locate_dotted_path(source)
            except ModuleNotFoundError:
                create_intermediate_modules(module_parts)
                spec = locate_dotted_path(source)

            source = Path(spec.origin)
            source.parent.mkdir(parents=True, exist_ok=True)
            original = source.read_text()

            module = ast.parse(original)

            names = {
                element.name
                for element in module.body if hasattr(element, 'name')
            }

            if fn_name not in names:
                print(f'Adding {fn_name!r} to module {source!s}...')
                fn_str = self.render('function.py', params=params)
                source.write_text(original + fn_str)

                did_create = True

        #  script task...
        else:
            path = Path(source)

            if not path.exists():
                if path.suffix in {'.py', '.sql', '.ipynb', '.R', '.Rmd'}:
                    # create parent folders if needed
                    source.parent.mkdir(parents=True, exist_ok=True)

                    content = self.render('task' + source.suffix,
                                          params=params)

                    print('Adding {}...'.format(source))
                    source.write_text(content)

                    did_create = True
                else:
                    print('Error: This command does not support adding '
                          'tasks with extension "{}", valid ones are '
                          '.py and .sql. Skipping {}'.format(
                              path.suffix, path))

        return did_create



from ploomber.util.dotted_path import DottedPath
from ploomber.exceptions import MissingClientError


class ProductWithClientMixin:
    """
    Adds the client property to a Product with the hierarchical resolution
    logic: Product -> Task -> DAG.clients. Product.client is only used
    for storing metadata
    """
    @property
    def client(self):
        if self._client is None:
            if self._task is None:
                raise ValueError('Cannot obtain client for this product, '
                                 'the constructor did not receive a client '
                                 'and this product has not been assigned '
                                 'to a DAG yet (cannot look up for clients in'
                                 ' dag.clients)')

            dag_client = self.task.dag.clients.get(type(self))

            if dag_client is None:
                raise MissingClientError(
                    f'{type(self).__name__} must be initialized with a client.'
                    ' Pass a client directly or set a DAG-level one')

            return dag_client

        if isinstance(self._client, DottedPath):
            # NOTE: should we test the output ot self._client()?
            # maybe check if it's a subclass of the Client abstract class
            self._client = self._client()

        return self._client


class SQLProductMixin:
    """
    Concrete SQL product classes must inherit from this one to signal
    they create SQL tables/views
    """
    pass



import sys
import os
import warnings
from pathlib import Path, WindowsPath
import importlib
from functools import wraps
import base64
import shutil
import inspect
from collections.abc import Iterable
from contextlib import contextmanager

from ploomber.exceptions import (CallbackSignatureError, CallbackCheckAborted,
                                 TaskRenderError)
from ploomber.util.dotted_path import DottedPath


def requires(pkgs, name=None, extra_msg=None, pip_names=None):
    """
    Check if packages were imported, raise ImportError with an appropriate
    message for missing ones

    Error message:
    a, b are required to use function. Install them by running pip install a b

    Parameters
    ----------
    pkgs : list
        The names of the packages required

    name
        The name of the module/function/class to show in the error message,
        if None, the decorated function __name__ attribute is used

    extra_msg
        Append this extra message to the end

    pip_names : list
        Pip package names to show in the suggested "pip install {name}"
        command, use it if different to the package name itself

    """
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            is_pkg_missing = [
                importlib.util.find_spec(pkg) is None for pkg in pkgs
            ]

            if any(is_pkg_missing):
                missing_pkgs = [
                    name for name, is_missing in zip(
                        pip_names or pkgs, is_pkg_missing) if is_missing
                ]

                fn_name = name or f.__name__

                raise ImportError(
                    _make_requires_error_message(missing_pkgs, fn_name,
                                                 extra_msg))

            return f(*args, **kwargs)

        return wrapper

    return decorator


def _make_requires_error_message(missing_pkgs, fn_name, extra_msg):
    names_str = ' '.join(repr(pkg) for pkg in missing_pkgs)

    error_msg = ('{} {} required to use {}. Install with: '
                 'pip install {}'.format(
                     names_str,
                     'is' if len(missing_pkgs) == 1 else 'are',
                     repr(fn_name),
                     names_str,
                 ))

    if extra_msg:
        error_msg += ('\n' + extra_msg)

    return error_msg


def check_mixed_envs(env_dependencies):
    # see: https://github.com/ploomber/soopervisor/issues/67
    env_dependencies = env_dependencies.split("\n")
    problematic_dependencies = [
        dep for dep in env_dependencies if ' @ file://' in dep
    ]
    if problematic_dependencies:
        warnings.warn("Found pip dependencies installed from local files.\n"
                      "This usually happens when using conda and pip"
                      " in the same environment, this will break "
                      "installation of new environments from the lock file."
                      " Problematic "
                      f"dependencies:\n{problematic_dependencies}")


def safe_remove(path):
    if path.exists():
        if path.is_file():
            path.unlink()
        else:
            shutil.rmtree(path)


def image_bytes2html(data):
    fig_base64 = base64.encodebytes(data)
    img = fig_base64.decode("utf-8")
    html = '<img src="data:image/png;base64,' + img + '"></img>'
    return html


def isiterable(obj):
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


def isiterable_not_str(obj):
    return isinstance(obj, Iterable) and not isinstance(obj, str)


# TODO: add more context to errors, which task and which hook?
def callback_check(fn, available, allow_default=True):
    """
    Check if a callback function signature requests available parameters

    Parameters
    ----------
    fn : callable
        Callable (e.g. a function) to check

    available : dict
        All available params

    allow_default : bool, optional
        Whether allow arguments with default values in "fn" or not

    Returns
    -------
    dict
        Dictionary with requested parameters

    Raises
    ------
    ploomber.exceptions.CallbackCheckAborted
        When passing a dotted path whose underlying function hasn't been
        imported
    ploomber.exceptions.CallbackSignatureError
        When fn does not have the required signature
    """
    # keep a copy of the original value because we'll modified it if this is
    # a DottedPath
    available_raw = available

    if isinstance(fn, DottedPath):
        available = {**fn._spec.get_kwargs(), **available}

        if fn.callable is None:
            raise CallbackCheckAborted(
                'Cannot check callback because function '
                'is a dotted path whose function has not been imported yet')
        else:
            fn = fn.callable

    parameters = inspect.signature(fn).parameters
    optional = {
        name
        for name, param in parameters.items()
        if param.default != inspect._empty
    }
    # not all functions have __name__ (e.g. partials)
    fn_name = getattr(fn, '__name__', fn)

    if optional and not allow_default:
        raise CallbackSignatureError('Callback functions cannot have '
                                     'parameters with default values, '
                                     'got: {} in "{}"'.format(
                                         optional, fn_name))

    required = {
        name
        for name, param in parameters.items()
        if param.default == inspect._empty
    }

    available_set = set(available)
    extra = required - available_set

    if extra:
        raise CallbackSignatureError('Callback function "{}" unknown '
                                     'parameter(s): {}, available ones are: '
                                     '{}'.format(fn_name, extra,
                                                 available_set))

    return {k: v for k, v in available_raw.items() if k in required}


def signature_check(fn, params, task_name):
    """
    Verify if the function signature used as source in a PythonCallable
    task matches available params
    """
    params = set(params)
    parameters = inspect.signature(fn).parameters

    required = {
        name
        for name, param in parameters.items()
        if param.default == inspect._empty
    }

    extra = params - set(parameters.keys())
    missing = set(required) - params

    errors = []

    if extra:
        msg = f'Got unexpected arguments: {sorted(extra)}'
        errors.append(msg)

    if missing:
        msg = f'Missing arguments: {sorted(missing)}'
        errors.append(msg)

    if 'upstream' in missing:
        errors.append('Verify this task declared upstream depedencies or '
                      'remove the "upstream" argument from the function')

    missing_except_upstream = sorted(missing - {'upstream'})

    if missing_except_upstream:
        errors.append(f'Pass {missing_except_upstream} in "params"')

    if extra or missing:
        msg = '. '.join(errors)
        # not all functions have __name__ (e.g. partials)
        fn_name = getattr(fn, '__name__', fn)
        raise TaskRenderError('Error rendering task "{}" initialized with '
                              'function "{}". {}'.format(
                                  task_name, fn_name, msg))

    return True


def call_with_dictionary(fn, kwargs):
    """
    Call a function by passing elements from a dictionary that appear in the
    function signature
    """
    parameters = inspect.signature(fn).parameters
    common = set(parameters) & set(kwargs)
    sub_kwargs = {k: kwargs[k] for k in common}
    return fn(**sub_kwargs)


def _make_iterable(o):
    if isinstance(o, Iterable) and not isinstance(o, str):
        return o
    elif o is None:
        return []
    else:
        return [o]


@contextmanager
def add_to_sys_path(path, chdir):
    cwd_old = os.getcwd()

    if path is not None:
        path = os.path.abspath(path)
        sys.path.insert(0, path)

        if chdir:
            os.chdir(path)

    try:
        yield
    finally:
        if path is not None:
            sys.path.remove(path)
            os.chdir(cwd_old)


def chdir_code(path):
    """
    Returns a string with valid code to chdir to the passed path
    """
    path = Path(path).resolve()

    if isinstance(path, WindowsPath):
        path = str(path).replace('\\', '\\\\')

    return f'os.chdir("{path}")'



from pathlib import Path, PurePosixPath
import datetime

import nbformat

from ploomber.tasks import PythonCallable


class JupyterTaskResource:
    def __init__(self, task, interactive, parent):
        self.task = task
        self.interactive = interactive
        self.path = '/'.join([parent, self.task.name])

    def to_model(self, content=False):
        last_modified = datetime.datetime.fromtimestamp(
            remove_line_number(self.task.source.loc).stat().st_mtime)
        return {
            'name': self.task.name,
            'type': 'notebook',
            'content': None if not content else self.interactive.to_nb(),
            'path': self.path,
            'writable': True,
            'created': datetime.datetime.now(),
            'last_modified': last_modified,
            'mimetype': None,
            'format': 'json' if content else None,
        }

    def __repr__(self):
        return (f'{type(self).__name__}(name={self.task.name!r}, '
                f'path={self.path})')


class JupyterDirectoryResource:
    def __init__(self, name, path):
        self.task_resources = dict()
        self.name = name
        self.path = path

    def __setitem__(self, key, value):
        self.task_resources[key] = value

    def get(self, key):
        return self.task_resources.get(key)

    def __iter__(self):
        for t in self.task_resources:
            yield t

    def __len__(self):
        return len(self.task_resources)

    def to_model(self, content=False):
        content = [
            t.to_model(content=False) for t in self.task_resources.values()
        ]
        created = min(c['created'] for c in content)
        last_modified = max(c['last_modified'] for c in content)
        return {
            'name': self.name,
            'path': self.path,
            'type': 'directory',
            'created': created,
            'last_modified': last_modified,
            'format': 'json',
            'mimetype': None,
            'content': content,
            'writable': True,
        }

    def __repr__(self):
        return (f'{type(self).__name__}(name={self.name!r}, '
                f'path={self.path}, '
                f'task_resources={self.task_resources!r})')


def as_jupyter_path(path):
    """
    Paths in Jupyter are delimited by / (even on Windows) and don't have
    trailing leading slashes. This function takes a platform-dependent
    path and converts it to a valid jupyter path

    Notes
    -----
    https://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html#api-paths
    """
    relative_path = Path(path).relative_to(Path('.').resolve())
    return relative_path.as_posix().strip('/')


def remove_line_number(path):
    """
    Takes a path/to/file:line path and returns path/to/file path object
    """
    parts = list(Path(path).parts)
    parts[-1] = parts[-1].split(':')[0]
    return Path(*parts)


class JupyterDAGManager:
    """
    Exposes PythonCallable tasks in a dag as Jupyter notebooks
    """
    def __init__(self, dag):
        self.resources = dict()

        for t in dag.values():
            if isinstance(t, PythonCallable):
                loc = remove_line_number(t.source.loc)
                name = loc.name + ' (functions)'
                key = str(PurePosixPath(as_jupyter_path(loc)).with_name(name))

                if key not in self.resources:
                    self.resources[key] = JupyterDirectoryResource(name=name,
                                                                   path=key)

                task_resource = JupyterTaskResource(
                    task=t, interactive=t._interactive_developer(), parent=key)
                self.resources[key][t.name] = task_resource
                # self.resources_keys.append(task_resource.path)
                self.resources[task_resource.path] = task_resource

        pairs = ((str(PurePosixPath(path).parent), res)
                 for path, res in self.resources.items())

        self.resources_by_root = dict()

        for parent, resource in pairs:
            if parent not in self.resources_by_root:
                self.resources_by_root[parent] = []

            self.resources_by_root[parent].append(resource)

    def __contains__(self, key):
        return key.strip('/') in self.resources

    def __getitem__(self, key):
        return self.resources[key]

    def __iter__(self):
        for resource in self.resources:
            yield resource

    def _get(self, path):
        path = path.strip('/')

        return self.resources.get(path)

    def get(self, path, content):
        """Get model located at path
        """
        resource = self._get(path)
        if resource:
            return resource.to_model(content)

    def get_by_parent(self, parent):
        parent = parent.strip('/')

        # jupyter represents the current folder with an empty string
        if parent == '':
            parent = '.'

        if parent in self.resources_by_root:
            return [m.to_model() for m in self.resources_by_root[parent]]
        else:
            return []

    def overwrite(self, model, path):
        """Overwrite a model back to the original function
        """
        resource = self._get(path)

        resource.interactive.overwrite(nbformat.from_dict(model['content']))

        return {
            'name': resource.task.name,
            'type': 'notebook',
            'path': path,
            'writable': True,
            'created': datetime.datetime.now(),
            'last_modified': datetime.datetime.now(),
            'content': None,
            'mimetype': 'text/x-python',
            'format': None,
        }

    def __repr__(self):
        return f'{type(self).__name__}({list(self.resources)})'



from typing import Union, List

from jinja2 import Template
from tabulate import tabulate

from ploomber.util.util import _make_iterable


def _duplicates_query(col, product):
    """Generate SQL code that counts number of duplicates
    """
    cols = ','.join(_make_iterable(col))

    return Template("""
    SELECT {{cols}}, COUNT(*) - 1 AS n_duplicates
    FROM {{product}}
    GROUP BY {{cols}}
    HAVING COUNT(*) > 1
    """).render(cols=cols, product=product)


def duplicates_in_column(client, col: Union[str, List[str]], product) -> bool:
    """Check if a column (or group of columns) has duplicated values

    Parameters
    ----------
    client
        Database client
    cols
        Column(s) to check
    product
        The relation to check

    Returns
    -------
    bool
        True if there are duplicates in the column(s). If passed more than
        one column, they are considered as a whole, not individually
    """
    sql = Template("""
    SELECT EXISTS(
        {{query}}
    )
    """).render(query=_duplicates_query(col, product))

    cur = client.connection.cursor()
    cur.execute(sql)
    output = bool(cur.fetchone()[0])
    cur.close()

    return output


def assert_no_duplicates_in_column(client,
                                   col: Union[str, List[str]],
                                   product,
                                   stats=False):
    """
    Assert if there are duplicates in a column (or group of columns). If there
    are duplicates, it raises an AssertionError with an error message showing
    some of the duplicated values

    Parameters
    ----------
    stats : bool, default=False
        Whether to show duplicates stats in the error message or not
    """
    duplicates_query = _duplicates_query(col, product)

    sql = Template("""
    {{query}}
    LIMIT 10
    """).render(query=duplicates_query)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = cur.fetchall()

    if len(output):
        names = [t[0] for t in cur.description]
        table = tabulate(output, headers=names)
        cols = ','.join(_make_iterable(col))

        sql_sample_rows = Template("""
        WITH duplicated AS (
            {{sql}}
        )
        SELECT t.*
        FROM {{product}} AS t
        JOIN duplicated
        USING ({{cols}})
        ORDER BY {{cols}}
        """).render(sql=sql, product=product, cols=cols)

        cur = client.connection.cursor()
        cur.execute(sql_sample_rows)
        output = cur.fetchall()
        cur.close()

        names = [t[0] for t in cur.description]
        table_sample = tabulate(output, headers=names)

        msg = f'Duplicates found.\n\n{table}\n\n{table_sample}'

        if stats:
            n_rows, n_unique, n_duplicates = duplicates_stats(
                client, col, product)

            msg += (f'\n\nNumber of rows: {n_rows:,}\n'
                    f'Number of unique values: {n_unique:,}\n'
                    f'Number of duplicates: {n_duplicates:,}')

        raise AssertionError(msg)


def duplicates_stats(client, col: Union[str, List[str]], product):
    """Get stats on rows with duplicated values

    Returns
    -------
    n_rows
        Number of rows in product
    n_unique
        Number of unique values (for selected columns) in product
    n_duplicates
        Number of rows with duplicated values (this is equal as the number
        of rows we'd have to drop to remove duplicates)
    """
    cols = ','.join(_make_iterable(col))

    # num of rows in product
    n_rows = _query(
        client,
        Template('SELECT COUNT(*) FROM {{product}}').render(product=product))

    # num of unique values (using all columns)
    n_unique = _query(
        client,
        Template('SELECT COUNT(DISTINCT({{cols}})) FROM {{product}}').render(
            product=product, cols=cols))

    sql_n_duplicates = Template("""
        WITH duplicated AS (
            {{sql}}
        )
        SELECT SUM(n_duplicates) FROM duplicated
        """).render(sql=_duplicates_query(col, product),
                    product=product,
                    cols=cols)

    # num of duplicated rows (number of rows we have to drop to remove all
    # duplicates)
    n_duplicates = _query(client, sql_n_duplicates)

    return n_rows, n_unique, n_duplicates


def _query(client, sql):
    cur = client.connection.cursor()
    cur.execute(sql)
    res = cur.fetchone()[0]
    cur.close()
    return res



from collections import abc


class AbstractDAG(abc.Mapping):
    """
    Abstract class for DAG-like objects. Its only purpose for now is in in the
    Task implementation, to check whether the `dag` argument is a DAG. In
    the future, we'll use this to define a DAG interface.
    """
    pass



import warnings
from functools import partial, wraps
from collections.abc import Mapping


class Assert:
    """
    An object to collect assertions and print results after they've all been
    evaluated

    Examples
    --------
    >>> from ploomber.validators import Assert
    >>> assert_ = Assert()
    >>> assert_(True, 'This wont be displayed')
    >>> assert_(False, 'This will be displayed')
    >>> assert_(False, 'This will also be displayed')
    >>> assert_.check() # doctest: +SKIP
    """
    def __init__(self):
        self.messages_error = []
        self.messages_warning = []

    def __call__(self, expression, error_message):
        if not expression:
            self.messages_error.append(error_message)

    def warn(self, expression, warning_message):
        if not expression:
            warnings.warn(warning_message)
            self.messages_warning.append(warning_message)

    def __len__(self):
        return len(self.messages_error)

    def __iter__(self):
        for msg in self.messages_error:
            yield msg

    def __repr__(self):
        return 'Assert oject with {} error messages'.format(len(self))

    def __str__(self):
        if not self.messages_error:
            str_ = 'No errors found'
        elif len(self.messages_error) == 1:
            str_ = '1 error found:\n{}'.format(self.messages_error[0])
        else:
            str_ = ('{} errors found:\n * {}'.format(
                len(self.messages_error), '\n * '.join(self.messages_error)))

        if len(self.messages_warning) == 1:
            str_ += '\n\n1 warning: {}'.format(self.messages_warning[0])
        elif len(self.messages_warning) > 1:
            str_ += ('\n\n{} warnings:\n * {}'.format(
                len(self.messages_warning),
                '\n * '.join(self.messages_warning)))

        return str_

    def check(self):
        """
        Raise AsserionError with all error messages if there is at least one
        error
        """
        if len(self):
            raise AssertionError(str(self))


def validator(fn):

    # TODO: verify fn signature

    @wraps(fn)
    def wrapped(**kwargs):
        if 'assert_' in kwargs:
            raise TypeError('Do not include the assert_ parameter in '
                            'validator functions')

        if 'data' in kwargs:
            raise TypeError('Do not include the data parameter in '
                            'validator functions')

        return partial(fn, **kwargs)

    return wrapped


@validator
def validate_schema(assert_,
                    data,
                    schema,
                    optional=None,
                    on_unexpected_cols='raise'):
    """Check if a data frame complies with a schema

    Parameters
    ----------
    data : pandas.DataFrame
        Data frame to test
    schema : list or dict
        List with column names (will only validate names)
        or dict with column names as keys, dtypes as values (will validate
        names and dtypes)
    optional : list, optional
        List of optional column names, no warns nor errors if they appear
    on_unexpected_cols : str, optional
        One of 'warn', 'raise' or None. If 'warn', it will warn on extra
        columns, if 'raise' it will raise an error, if None it will completely
        ignore extra columns
    """
    if on_unexpected_cols not in {'warn', 'raise', None}:
        raise ValueError("'on_unexpected_cols' must be one of 'warn', 'raise' "
                         "or None")

    optional = optional or {}
    cols = set(data.columns)
    expected = set(schema)

    missing = expected - cols
    unexpected = cols - expected - set(optional)

    msg = '(validate_schema) Missing columns {missing}.'.format(
        missing=missing)
    assert_(not missing, msg)

    if on_unexpected_cols is not None:
        msg = ('(validate_schema) Unexpected columns {unexpected}'.format(
            unexpected=unexpected))
        caller = assert_ if on_unexpected_cols == 'raise' else assert_.warn
        caller(not unexpected, msg)

    # if passing a mapping, schema is validated (even for optional columns)
    for schema_to_validate in [schema, optional]:
        if isinstance(schema_to_validate, Mapping):
            # validate column types (as many as you can)
            dtypes = data.dtypes.astype(str).to_dict()

            for name, dtype in dtypes.items():
                expected = schema_to_validate.get(name)

                if expected is not None:
                    msg = (
                        '(validate_schema) Wrong dtype for column "{name}". '
                        'Expected: "{expected}". Got: "{dtype}"'.format(
                            name=name, expected=expected, dtype=dtype))
                    assert_(dtype == expected, msg)

    return assert_


@validator
def validate_values(assert_, data, values):
    data_cols = data.columns

    for column, (kind, params) in values.items():
        if column not in data_cols:
            assert_.warn(False,
                         ('(validate_values) Declared spec for column "{}" but'
                          ' it does not appear in the data').format(column))
        elif kind == 'unique':
            expected = set(params)
            unique = set(data[column].unique())
            msg = ('(validate_values) Expected unique values of  "{}" to be a'
                   ' subset of {}, got: {}'.format(column, expected, unique))
            assert_(expected >= unique, msg)
        elif kind == 'range':
            if len(params) != 2:
                raise ValueError('If kind is range, you must provide two '
                                 'values, got {}'.format(params))
            min_expected, max_expected = params
            min_ = data[column].min()
            max_ = data[column].max()
            msg = ('(validate_values) Expected range of "{}" to be [{}, {}], '
                   'got [{}, {}]'.format(column, min_expected, max_expected,
                                         min_, max_))
            assert_(min_expected <= min_ and max_ <= max_expected, msg)
        else:
            raise ValueError('Got invalid kind, must be "unique" or "range"')


def data_frame_validator(df, validators):
    """

    Examples
    --------
    >>> from ploomber.validators import data_frame_validator
    >>> from ploomber.validators import validate_schema, validate_values
    >>> import pandas as pd
    >>> import numpy as np
    >>> df = pd.DataFrame({'x': np.random.rand(3), 'y': np.random.rand(3),
    ...                    'z': [0, 1, 2], 'i': ['a', 'b', 'c']})
    >>> data_frame_validator(df,
    ...                     [validate_schema(schema={'x': 'int', 'z': 'int'}),
    ...                      validate_values(values={'z': ('range', (0, 1)),
    ...                                              'i': ('unique', {'a'}),
    ...                                              'j': ('unique', {'b'})}
    ...                                      )]) # doctest: +SKIP
    """
    assert_ = Assert()

    for validator in validators:
        validator(assert_=assert_, data=df)

    if len(assert_):
        raise AssertionError('Data frame validation failed. ' + str(assert_))

    return True



import argparse
import json
import shutil
from pathlib import Path
import stat

import click

from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import command_endpoint
from ploomber.telemetry import telemetry
from ploomber.sources.notebooksource import recursive_update
from ploomber.exceptions import BaseException


def _call_in_source(dag, method_name, message, kwargs=None):
    """
    Execute method on each task.source in dag, passing kwargs
    """
    kwargs = kwargs or {}
    files = []
    results = []

    for task in dag.values():
        try:
            method = getattr(task.source, method_name)
        except AttributeError:
            pass
        else:
            results.append(method(**kwargs))
            files.append(str(task.source._path))

    files_ = '\n'.join((f'    {f}' for f in files))
    click.echo(f'{message}:\n{files_}')

    return results


def _install_hook(path_to_hook, content, entry_point):
    """
    Install a git hook script at the given path
    """
    if path_to_hook.exists():
        raise RuntimeError(
            'hook already exists '
            f'at {path_to_hook}. Run: "ploomber nb -u" to uninstall the '
            'existing hook and try again')

    path_to_hook.write_text(content.format(entry_point=entry_point))
    # make the file executable
    path_to_hook.chmod(path_to_hook.stat().st_mode | stat.S_IEXEC)


def _delete_hook(path):
    """Delete a git hook at the given path
    """
    if path.exists():
        if path.is_file():
            path.unlink()
        else:
            # in the remote case that it's a directory
            shutil.rmtree(path)

    click.echo(f'Deleted hook located at {path}')


pre_commit_hook = """
# !/usr/bin/env bash
# Automatically generated pre-commit hook to remove the injected cell in
# scripts and notebook tasks

# remove injected cells
ploomber nb --entry-point {entry_point} --remove

# re-add files
git add $(git diff --name-only --cached)
"""

post_commit_hook = """
# !/usr/bin/env bash
# Automatically generated post-commit hook to add the injected cell in
# scripts and notebook tasks

# inject cells
ploomber nb --entry-point {entry_point} --inject
"""

# taken from https://github.com/mwouts/jupytext/blob/main/README.md#install
_jupyterlab_default_settings_overrides = """
{
  "@jupyterlab/docmanager-extension:plugin": {
    "defaultViewers": {
      "markdown": "Jupytext Notebook",
      "myst": "Jupytext Notebook",
      "r-markdown": "Jupytext Notebook",
      "quarto": "Jupytext Notebook",
      "julia": "Jupytext Notebook",
      "python": "Jupytext Notebook",
      "r": "Jupytext Notebook"
    }
  }
}
"""


def _py_with_single_click_enable():
    """
    Writes ~/.jupyterlab/labconfig/default_setting_overrides.json to enable
    opening .py files as notebooks with a single click. If the secion already
    exists, it overrides its value
    """
    parent = Path('~/.jupyter', 'labconfig').expanduser()
    path = parent / 'default_setting_overrides.json'

    if path.exists():
        target = json.loads(path.read_text())
    else:
        target = {}

    recursive_update(target,
                     json.loads(_jupyterlab_default_settings_overrides))

    click.echo(f'Overriding JupyterLab defaults at: {str(path)}')
    parent.mkdir(exist_ok=True, parents=True)
    path.write_text(json.dumps(target))
    click.secho(
        'Done. You can now open .py and other formats in JupyterLab '
        'with a single click. You may need to reload JupyterLab',
        fg='green')


def _py_with_single_click_disable():
    """
    Opens ~/.jupyterlab/labconfig/default_setting_overrides.json and deletes
    the value in
    ['@jupyterlab/docmanager-extension:plugin'][''defaultViewers'], if any
    """
    parent = Path('~/.jupyter', 'labconfig')
    target = (parent / 'default_setting_overrides.json').expanduser()

    if target.exists():
        content = json.loads(target.read_text())
        key1 = '@jupyterlab/docmanager-extension:plugin'
        key2 = 'defaultViewers'

        if content.get(key1, {}).get(key2):
            del content[key1][key2]

            if key1 in content and not content.get(key1):
                del content[key1]

            Path(target).write_text(json.dumps(content))

    click.secho(
        'Done. Disabled opening .py files and other formats in JupyterLab '
        'with a single click. You may need to reload JupyterLab',
        fg='yellow')


_description = """Manage scripts and notebooks


Inject cell in all scripts and notebooks:

$ ploomber nb -i

Enable one-click to open .py as notebooks in JupyterLab:

$ ploomber nb -S

Re-format .ipynb notebooks as .py files with the percent format:

$ ploomber nb -f py:percent

Re-format .py files as .ipynb notebooks:

$ ploomber nb -f ipynb
"""


# TODO: --log, --log-file should not appear as options
@command_endpoint
@telemetry.log_call('nb')
def main():
    parser = CustomParser(
        description=_description,
        prog='ploomber nb',
        # required for the --help text to keep line breaks
        formatter_class=argparse.RawTextHelpFormatter)

    with parser:
        # The next options do not require a valid entry point

        # opening .py files as notebooks in JupyterLab with a single click
        single_click = parser.add_mutually_exclusive_group()
        single_click.add_argument(
            '--single-click',
            '-S',
            action='store_true',
            help=('Override JupyterLab defaults to open '
                  'scripts as notebook with a single click'))
        single_click.add_argument(
            '--single-click-disable',
            '-d',
            action='store_true',
            help=('Disables opening scripts as notebook with a single '
                  'click in JupyterLab'))

        # install/uninstall hook
        hook = parser.add_mutually_exclusive_group()
        hook.add_argument('--install-hook',
                          '-I',
                          action='store_true',
                          help='Install git pre-commit hook')
        hook.add_argument('--uninstall-hook',
                          '-u',
                          action='store_true',
                          help='Uninstall git pre-commit hook')

        # The next options require a valid entry point

        # inject/remove cell
        cell = parser.add_mutually_exclusive_group()
        cell.add_argument('--inject',
                          '-i',
                          action='store_true',
                          help='Inject cell to all script/notebook tasks')
        cell.add_argument(
            '--remove',
            '-r',
            action='store_true',
            help='Remove injected cell in all script/notebook tasks')

        # re-format
        parser.add_argument('--format',
                            '-f',
                            help='Re-format script/notebook tasks '
                            '(values: "py:percent" and "ipynb")')

        # pair scripts and nbs
        parser.add_argument('--pair',
                            '-p',
                            help='Pair scripts with ipynb files')

        # sync scripts and nbs
        parser.add_argument('--sync',
                            '-s',
                            action='store_true',
                            help='Sync scripts with ipynb files')

    loading_error = None

    # commands that need an entry point to work
    needs_entry_point = {'format', 'inject', 'remove', 'sync', 'pair'}

    args_ = parser.parse_args()

    if any(getattr(args_, arg) for arg in needs_entry_point):
        try:
            dag, args = parser.load_from_entry_point_arg()
        except Exception as e:
            loading_error = e
        else:
            dag.render(show_progress=False)

        if loading_error:
            raise BaseException('Could not run nb command: the DAG '
                                'failed to load') from loading_error
    else:
        dag = None
        args = args_

    # options that do not need a DAG

    if args.single_click:
        _py_with_single_click_enable()

    if args.single_click_disable:
        _py_with_single_click_disable()

    if args.install_hook:
        if not Path('.git').is_dir():
            raise NotADirectoryError(
                'Expected a .git/ directory in the current working '
                'directory. Run this from the repository root directory.')

        parent = Path('.git', 'hooks')
        parent.mkdir(exist_ok=True)

        # pre-commit: remove injected cells
        _install_hook(parent / 'pre-commit', pre_commit_hook, args.entry_point)
        click.echo('Successfully installed pre-commit git hook')

        # post-commit: inject cells
        _install_hook(parent / 'post-commit', post_commit_hook,
                      args.entry_point)
        click.echo('Successfully installed post-commit git hook')

    if args.uninstall_hook:
        _delete_hook(Path('.git', 'hooks', 'pre-commit'))
        _delete_hook(Path('.git', 'hooks', 'post-commit'))

    # options that need a valid DAG

    if args.format:
        new_paths = [
            str(p) for p in _call_in_source(
                dag,
                'format',
                'Formatted notebooks',
                dict(fmt=args.format),
            ) if p is not None
        ]

        if len(new_paths):
            click.echo('Extension changed for the following '
                       f'tasks: {", ".join(new_paths)}. Update your '
                       'pipeline declaration.')

    if args.inject:
        _call_in_source(
            dag,
            'save_injected_cell',
            'Injected cell',
            dict(),
        )

        click.secho(
            'Finished cell injection. Re-run this command if your '
            'pipeline.yaml changes.',
            fg='green')

    if args.remove:
        _call_in_source(
            dag,
            'remove_injected_cell',
            'Removed injected cell',
            dict(),
        )

    if args.sync:
        # maybe its more efficient to pass all notebook paths at once?
        _call_in_source(dag, 'sync', 'Synced notebooks')

    # can pair give trouble if we're reformatting?
    if args.pair:
        _call_in_source(
            dag,
            'pair',
            'Paired notebooks',
            dict(base_path=args.pair),
        )
        click.echo(f'Finished pairing notebooks. Tip: add {args.pair!r} to '
                   'your .gitignore to keep your repository clean')



class TaskFactory:
    """Utility class for reducing boilerplate code
    """
    def __init__(self, task_class, product_class, dag):
        self.task_class = task_class
        self.product_class = product_class
        self.dag = dag

    def make(self, task_arg, product_arg, name, params=None):
        product = self.product_class(product_arg)
        return self.task_class(task_arg,
                               product=product,
                               dag=self.dag,
                               name=name,
                               params=params)



import warnings

import sqlparse

from ploomber import products
from ploomber.static_analysis.abc import Extractor
from ploomber.static_analysis.jinja import JinjaExtractor


class SQLExtractor(Extractor):
    """Extract upstream and product from SQL templates

    Parameters
    ----------
    code : str or Placeholder
        SQL code
    """
    def __init__(self, code):
        self._jinja_extractor = JinjaExtractor(code)
        self._product = None
        self._extracted_product = False

    def extract_upstream(self):
        """Extract upstream keys used in a templated SQL script
        """
        return self._jinja_extractor.find_variable_access(variable='upstream')

    def extract_product(self, raise_if_none=True):
        """
        Extract an object from a SQL template that defines as product variable:

        {% set product = SOME_CLASS(...) %}

        Where SOME_CLASS is a class defined in ploomber.products. If no product
        variable is defined, returns None
        """
        product = self._jinja_extractor.find_variable_assignment(
            variable='product')

        if product is None:
            if raise_if_none:
                code = self._jinja_extractor.get_code_as_str()
                raise ValueError(
                    f"Couldn't extract 'product' from code: {code!r}")
        else:
            # validate product
            try:
                # get the class name used
                class_ = getattr(products, product.node.name)
                # get the arg passed to the class
                arg = product.args[0].as_const()
                # try to initialize object
                return class_(arg)
            except Exception as e:
                exc = ValueError("Found a variable named 'product' in "
                                 "code: {} but it does not appear to "
                                 "be a valid SQL product, verify it ".format(
                                     self._jinja_extractor.code))
                raise exc from e


def _quoted_with(identifier, char):
    return identifier.startswith(char) and identifier.endswith(char)


def _normalize(identifier):
    """
    Normalize a SQL identifier. Given that different SQL implementations have
    different rules, we will implement logic based on PostgreSQL. Double
    quotes make an identifier case sensitive, unquoted are forced to lower
    case. MySQL on the other hand, depends on the file system, furthermore
    the quoting identifier character can also be a backtick.

    Notes
    -----
    PostgreSQL - Section 4.1.1 mentions quoted identifiers:
    https://www.postgresql.org/docs/9.1/sql-syntax-lexical.html

    MySQL - Section 9.2:
    https://dev.mysql.com/doc/refman/8.0/en/identifiers.html
    https://dev.mysql.com/doc/refman/8.0/en/identifier-case-sensitivity.html
    """
    # does this cover all use cases?
    if identifier is None:
        return None
    elif _quoted_with(identifier, '"'):
        return identifier.replace('"', '')
    else:
        return identifier.lower()


class ParsedSQLRelation:
    """
    This is similar to the SQLRelationPlaceholder, it enables operations
    such as comparisons. Not using SQLRelationPlaceholder directly to
    avoid complex imports
    """
    def __init__(self, schema, name, kind):
        self.schema = schema
        self.name = name
        self.kind = kind

    def __eq__(self, other):
        return (_normalize(self.schema) == _normalize(other.schema)
                and _normalize(self.name) == _normalize(other.name)
                and _normalize(self.kind) == _normalize(other.kind))

    def __str__(self):
        if self.schema is None:
            return self.name
        else:
            return '{}.{}'.format(self.schema, self.name)

    def __repr__(self):
        raw_repr = (self.name,
                    self.kind) if self.schema is None else (self.schema,
                                                            self.name,
                                                            self.kind)
        return f'{type(self).__name__}({raw_repr!r})'

    def __hash__(self):
        return hash((self.schema, self.name, self.kind))


def name_from_create_statement(statement):
    # NOTE: is there a better way to look for errors?
    errors = [
        t.ttype for t in statement.tokens if 'Token.Error' in str(t.ttype)
    ]

    if any(errors):
        warnings.warn('Failed to parse statement: {}'.format(str(statement)))
    else:
        # after removing whitespace this should be
        # [CREATE, 'TABLE|VIEW', 'IF EXISTS'?, 'IDENTIFIER']
        elements = [t for t in statement.tokens if not t.is_whitespace]

        # get the first identifier and extract tokens
        tokens = [
            e for e in elements if isinstance(e, sqlparse.sql.Identifier)
        ][0].tokens

        if len(tokens) == 1:
            return ParsedSQLRelation(schema=None,
                                     name=tokens[0].value,
                                     kind=str(elements[1]))

        else:
            return ParsedSQLRelation(schema=tokens[0].value,
                                     name=tokens[2].value,
                                     kind=str(elements[1]))


def created_relations(sql, split_source=None):
    """
    Determine the name of the relations that will be created if we run a given
    sql script

    Parameters
    ----------
    sql : str
        SQL code

    split_source : str
        The character used to delimit multiple commands. If not None, each
        command is analyzed separately and assumed to be executed in order
    """
    # the split_source argument makes the database client able to execute
    # multiple statements at once even if the driver does not allow it, by
    # splitting by a character (usually ';') and executing one task at a time.
    # there is one special case where one might want to use a different
    # character, this will cause sqlparse.parse to fail so we have to split
    # and re-join using ';' which is an acceptable delimiter in SQL
    if split_source:
        sql = ';'.join(sql.split(split_source))

    sql = sqlparse.format(sql,
                          keyword_case='lower',
                          identifier_case='lower',
                          strip_comments=True)
    statements = sqlparse.parse(sql)

    # get DROP and CREATE statements along with their indexes, which give
    # the position
    drop_idx = {
        name_from_create_statement(s): idx
        for idx, s in enumerate(statements) if s.get_type() == 'DROP'
    }
    create_idx = {
        name_from_create_statement(s): idx
        for idx, s in enumerate(statements) if s.get_type() == 'CREATE'
    }

    relations = []

    # add a relation to the final list if...
    for relation, idx in create_idx.items():
        # there isn't a DROP statement with the same name
        # OR
        # there is a DROP statement but comes before the CREATE
        # e.g. DROP TABLE x; CREATE TABLE x AS ...
        if relation not in drop_idx or drop_idx[relation] < idx:
            relations.append(relation)

    return relations



import glob
import abc
from pathlib import PurePosixPath, Path

from ploomber.clients.storage.util import _resolve


class AbstractStorageClient(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def download(self, local, destination=None):
        """
        Download remote copy of a given local path. Local may be a file
        or a folder (all contents downloaded).

        Parameters
        ----------
        local
            Path to local file or folder whose remote copy will be downloaded

        destination
            Download location. If None, overwrites local copy
        """
        pass

    def upload(self, local):
        """Upload file or folder from a local path by calling _upload as needed

        Parameters
        ----------
        local
            Path to local file or folder to upload
        """
        if Path(local).is_dir():
            for f in glob.iglob(str(Path(local, '**')), recursive=True):
                if Path(f).is_file():
                    self._upload(f)
        else:
            self._upload(local)

    @abc.abstractmethod
    def _download(self, local, destination):
        """Download a single file
        """
        pass

    @abc.abstractmethod
    def _upload(self, local):
        """Upload a single file
        """
        pass

    @abc.abstractmethod
    def _is_file(self, remote):
        """Check if path to remote is file
        """
        pass

    @abc.abstractmethod
    def _is_dir(self, remote):
        """Check if path to remote is a directory
        """
        pass

    def _remote_path(self, local):
        """
        Given a local path, compute the remote path where the file will be
        stored.

        1. Obtain the absolute project root (``/path/to/project``)
        2. Get the local absolute path (``/path/to/project/out/data.csv``)
        3. Compute the relative path (``out/data.csv``)
        4. Prefix the relative path with the ``parent`` argument
        (passed to the Client constructor) (``path/to/parent/out/data.csv``)
        """
        relative = _resolve(local).relative_to(self._path_to_project_root)
        return str(PurePosixPath(self._parent, *relative.parts))

    def _remote_exists(self, local):
        remote = self._remote_path(local)
        is_file = self._is_file(remote)

        if is_file:
            return True

        return self._is_dir(remote)

    def close(self):
        # required to comply with the Client API
        pass



from pathlib import Path

import click

from ploomber.scaffold.scaffoldloader import ScaffoldLoader
from ploomber.util.util import add_to_sys_path
from ploomber.util import loader
from ploomber.exceptions import DAGSpecInvalidError


def load_dag():
    # setting lazy_import to true causes sources to be returned as paths,
    # instead of placeholders
    try:
        return loader._default_spec_load(lazy_import='skip')
    except DAGSpecInvalidError:
        return None


def add(spec, path_to_spec):
    """Add scaffold templates for tasks whose source does not exist

    Parameters
    ----------
    spec : DAGSpec
        The spec to inspect to create missing task.source

    path_to_spec : str
        Path to the spec, only used to emit messages to the console
    """
    loader = ScaffoldLoader()

    # TODO: when the dag has a source loader, the argument passed to
    # ploomber_add should take that into account to place the new file
    # in the appropriate location (instead of doing it relative to
    # pipeline.yaml)

    # TODO: raise an error if the location is inside the site-packages folder

    # NOTE: lazy loading freom source loader will giev errors because
    # initializing a source with a path only, loses the information from the
    # jinja environment to make macros workj. I have to test this. the best
    # solution is to add a lazy_load param to Placeholder, so it can be
    # initialized with a path for a file that does not exist

    if path_to_spec:
        click.echo(f'Found spec at {str(path_to_spec)!r}')

        n = 0

        # make sure current working dir is in the path, otherwise we might not
        # be able to import the PythonCallable functions, which we need to do
        # to locate the modules
        path_to_parent = str(Path(path_to_spec).resolve().parent)

        with add_to_sys_path(path_to_parent, chdir=False):
            for task in spec['tasks']:
                did_create = loader.create(source=task['source'],
                                           params=spec['meta'],
                                           class_=task['class'])
                n += int(did_create)

        if not n:
            click.echo(f'All tasks sources declared in {str(path_to_spec)!r} '
                       'exist, nothing was created.')
        else:
            click.echo(f'Created {n} new task sources.')

    else:
        click.echo('Error: No pipeline.yaml spec found...')



import base64
import json


class Base64Serializer:
    @staticmethod
    def serialize(metadata):
        json_str = json.dumps(metadata)
        return base64.encodebytes(json_str.encode('utf-8')).decode('utf-8')

    @staticmethod
    def deserialize(metadata_str):
        bytes_ = metadata_str.encode('utf-8')
        metadata = json.loads(base64.decodebytes(bytes_).decode('utf-8'))
        return metadata



import time
from pathlib import Path
import os
from difflib import get_close_matches
import sys

from ploomber_scaffold import scaffold as scaffold_project

import click

CLICK_VERSION = int(click.__version__[0])
# NOTE: package_name was introduced in version 8
VERSION_KWARGS = dict(
    package_name='ploomber') if CLICK_VERSION >= 8 else dict()


def _suggest_command(name: str, options):
    if not name or name in options:
        return None

    name = name.lower()

    mapping = {
        'run': 'build',
        'execute': 'build',
    }

    if name in mapping:
        return mapping[name]

    close_commands = get_close_matches(name, options)

    if close_commands:
        return close_commands[0]
    else:
        return None


@click.group()
@click.version_option(**VERSION_KWARGS)
def cli():
    """
    Ploomber

    Need help? https://ploomber.io/community

    Download an example to the mypipeline/ directory:

    $ ploomber examples -n templates/ml-basic -o mypipeline

    Create a new project in a myproject/ directory:

    $ ploomber scaffold myproject

    """
    pass  # pragma: no cover


@cli.command()
@click.option(
    '--conda/--pip',
    '-c/-p',
    is_flag=True,
    default=None,
    help='Use environment.yaml/requirements.txt for dependencies',
)
@click.option(
    '--package',
    '-P',
    is_flag=True,
    help='Use package template (creates setup.py)',
)
@click.option(
    '--empty',
    '-E',
    is_flag=True,
    help='Create a pipeline.yaml with no tasks',
)
@click.option(
    '--entry-point',
    '-e',
    default=None,
    help='Entry point to add tasks. Invalid if other flags present',
)
@click.argument('name', required=False)
def scaffold(name, conda, package, entry_point, empty):
    """
    Create a new project and task source files

    Step 1. Create new project

      $ ploomber scaffold myproject

      $ cd myproject

    Step 2. Add tasks to the pipeline.yaml file

    Step 3. Create source files

      $ ploomber scaffold

    Need help? https://ploomber.io/community
    """
    from ploomber import scaffold as _scaffold
    from ploomber.telemetry import telemetry

    template = '-e/--entry-point is not compatible with {flag}'
    user_passed_name = name is not None

    if entry_point and name:
        err = '-e/--entry-point is not compatible with the "name" argument'
        telemetry.log_api("scaffold_error",
                          metadata={
                              'type': 'entry_and_name',
                              'exception': err,
                              'argv': sys.argv
                          })
        raise click.ClickException(err)

    if entry_point and conda:
        err = template.format(flag='--conda')
        telemetry.log_api("scaffold_error",
                          metadata={
                              'type': 'entry_and_conda_flag',
                              'exception': err,
                              'argv': sys.argv
                          })
        raise click.ClickException(err)

    if entry_point and package:
        err = template.format(flag='--package')
        telemetry.log_api("scaffold_error",
                          metadata={
                              'type': 'entry_and_package_flag',
                              'exception': err,
                              'argv': sys.argv
                          })
        raise click.ClickException(err)

    if entry_point and empty:
        err = template.format(flag='--empty')
        telemetry.log_api("scaffold_error",
                          metadata={
                              'type': 'entry_and_empty_flag',
                              'exception': err,
                              'argv': sys.argv
                          })
        raise click.ClickException(err)

    # try to load a dag by looking in default places
    if entry_point is None:
        loaded = _scaffold.load_dag()
    else:
        from ploomber.spec import DAGSpec

        try:
            loaded = (
                DAGSpec(entry_point, lazy_import='skip'),
                Path(entry_point).parent,
                Path(entry_point),
            )
        except Exception as e:
            telemetry.log_api("scaffold_error",
                              metadata={
                                  'type': 'dag_load_failed',
                                  'exception': str(e),
                                  'argv': sys.argv
                              })
            raise click.ClickException(e) from e

    if loaded:
        if user_passed_name:
            click.secho(
                "The 'name' positional argument is "
                "only valid for creating new projects, ignoring...",
                fg='yellow')

        # existing pipeline, add tasks
        spec, _, path_to_spec = loaded
        _scaffold.add(spec, path_to_spec)

        telemetry.log_api("ploomber_scaffold",
                          metadata={
                              'type': 'add_task',
                              'argv': sys.argv,
                              'dag': loaded,
                          })
    else:
        # no pipeline, create base project
        telemetry.log_api("ploomber_scaffold",
                          metadata={
                              'type': 'base_project',
                              'argv': sys.argv
                          })
        scaffold_project.cli(project_path=name,
                             conda=conda,
                             package=package,
                             empty=empty)


@cli.command()
@click.option('--use-lock/--no-use-lock',
              '-l/-L',
              default=None,
              help=('Use lock/regular files. If not present, uses any '
                    'of them, prioritizing lock files'))
@click.option('--create-env',
              '-e',
              is_flag=True,
              help=('Create a new environment, otherwise install in the '
                    'current environment'))
@click.option(
    '--use-venv',
    '-v',
    is_flag=True,
    help='Use Python\'s venv module (ignoring conda if installed)',
)
def install(use_lock, create_env, use_venv):
    """
    Install dependencies
    """
    from ploomber import cli as cli_module

    cli_module.install.main(use_lock=use_lock,
                            create_env=create_env,
                            use_venv=use_venv)


@cli.command()
@click.option('-n', '--name', help='Example to download', default=None)
@click.option('-f', '--force', help='Force examples download', is_flag=True)
@click.option('-o', '--output', help='Target directory', default=None)
@click.option('-b', '--branch', help='Git branch to use.', default=None)
def examples(name, force, branch, output):
    """
    Download examples

    Step 1. List examples

      $ ploomber examples


    Step 2. Download an example

      $ ploomber examples -n templates/ml-basic -o my-pipeline


    Need help? https://ploomber.io/community
    """
    click.echo('Loading examples...')

    from ploomber.telemetry import telemetry
    from ploomber import cli as cli_module

    try:
        cli_module.examples.main(name=name,
                                 force=force,
                                 branch=branch,
                                 output=output)
    except click.ClickException:
        raise
    except Exception as e:
        telemetry.log_api("examples_error",
                          metadata={
                              'type': 'runtime_error',
                              'exception': str(e),
                              'argv': sys.argv
                          })
        raise RuntimeError(
            'An error happened when executing the examples command. Check out '
            'the full error message for details. Downloading the examples '
            'again or upgrading Ploomber may fix the '
            'issue.\nDownload: ploomber examples -f\n'
            'Update: pip install ploomber -U\n'
            'Update [conda]: conda update ploomber -c conda-forge') from e


def _exit_with_error_message(msg):
    click.echo(msg, err=True)
    sys.exit(2)


def cmd_router():
    """CLI entry point
    """
    cmd_name = None if len(sys.argv) < 2 else sys.argv[1]

    # These are parsing dynamic parameters and that's why we're isolating it.
    custom = ['build', 'plot', 'task', 'report', 'interact', 'status', 'nb']

    # users may attempt to run execute/run, suggest to use build instead
    # users may make typos when running one of the commands
    # suggest correct spelling on obvious typos

    if cmd_name in custom:
        click.echo('Loading pipeline...')

        from ploomber import cli as cli_module

        # NOTE: we don't use the argument here, it is parsed by _main
        # pop the second element ('entry') to make the CLI behave as expected
        sys.argv.pop(1)
        # Add the current working directory, this is done automatically when
        # calling "python -m ploomber.build" but not here ("ploomber build")
        sys.path.insert(0, os.path.abspath('.'))

        custom = {
            'build': cli_module.build.main,
            'plot': cli_module.plot.main,
            'task': cli_module.task.main,
            'report': cli_module.report.main,
            'interact': cli_module.interact.main,
            'status': cli_module.status.main,
            'nb': cli_module.nb.main,
        }

        fn = custom[cmd_name]
        fn()
    else:
        suggestion = _suggest_command(cmd_name, cli.commands.keys())

        # Set nested command keys based on command option
        nested_cmds = {'cloud': cloud.commands.keys()}

        # Evaluate nested command and provide suggested command if applicable
        if cmd_name in nested_cmds:
            nested_cmd_name = None if len(sys.argv) < 3 else sys.argv[2]
            nested_suggestion = _suggest_command(nested_cmd_name,
                                                 nested_cmds[cmd_name])
        else:
            nested_suggestion = None

        if suggestion:
            _exit_with_error_message(
                "Try 'ploomber --help' for help.\n\n"
                f"Error: {cmd_name!r} is not a valid command."
                f" Did you mean {suggestion!r}?")
        elif nested_suggestion:
            cmds = f'{cmd_name} {nested_cmd_name}'
            suggestion = f'{cmd_name} {nested_suggestion}'
            suggested_help = f'ploomber {cmd_name} --help'

            _exit_with_error_message(f"Try {suggested_help!r} for help.\n\n"
                                     f"Error: {cmds!r} is not a valid command."
                                     f" Did you mean {suggestion!r}?")
        else:
            cli()


# the commands below are handled by the router,
# those are a place holder to show up in ploomber --help
@cli.command()
def build():
    """Build pipeline
    """
    pass  # pragma: no cover


@cli.command()
def status():
    """Show pipeline status
    """
    pass  # pragma: no cover


@cli.command()
def plot():
    """Plot pipeline
    """
    pass  # pragma: no cover


@cli.command()
def task():
    """Interact with specific tasks
    """
    pass  # pragma: no cover


@cli.command()
def report():
    """Generate pipeline report
    """
    pass  # pragma: no cover


@cli.command()
def interact():
    """Interact with your pipeline (REPL)
    """
    pass  # pragma: no cover


@cli.command()
def nb():
    """Manage scripts and notebooks
    """
    pass  # pragma: no cover


@cli.group()
def cloud():
    """Manage Ploomber Cloud
    """
    pass  # pragma: no cover


@cloud.command()
@click.argument('user_key', required=True)
def set_key(user_key):
    """Set API key

    It validates it's a valid key. If the key is valid, it stores it in the
    user config.yaml file.
    """
    from ploomber import cli as cli_module
    cli_module.cloud.set_key(user_key)


@cloud.command()
def get_key():
    """Return API key

    It retrieves it from the user config.yaml file. Returns the key if it
    exists
    """
    from ploomber import cli as cli_module

    key = cli_module.cloud.get_key()

    if key:
        click.echo(f'This is your cloud API key: {key}')
    else:
        click.echo('No cloud API key was found.')


@cloud.command()
@click.argument('pipeline_id', default=None, required=False)
@click.option('-v', '--verbose', default=False, is_flag=True, required=False)
def get_pipelines(pipeline_id, verbose):
    """Get pipeline status

    Specify a pipeline_id to get it's state, when not specified, pulls all of
    the pipelines. You can pass "latest", instead of pipeline_id to get the
    latest. To get a detailed view pass the verbose flag.
    Returns a list of pipelines is succeeds, an Error string if fails.
    """
    from ploomber.table import Table
    from ploomber import cli as cli_module

    pipeline = cli_module.cloud.get_pipeline(pipeline_id, verbose)
    if isinstance(pipeline, list) and pipeline:
        pipeline = Table.from_dicts(pipeline, complete_keys=True)
    print(pipeline)


@cloud.command()
@click.argument('pipeline_id', required=True)
@click.argument('status', required=True)
@click.argument('log', default=None, required=False)
@click.argument('dag', default=None, required=False)
@click.argument('pipeline_name', default=None, required=False)
def write_pipeline(pipeline_id, status, log, pipeline_name, dag):
    """Write a pipeline

    Pipeline_id & status are required. You can also add logs and a pipeline
    name. Returns a string with pipeline id if succeeds, an Error string if
    fails.
    """
    from ploomber import cli as cli_module

    print(
        cli_module.cloud.write_pipeline(pipeline_id, status, log,
                                        pipeline_name, dag))


@cloud.command()
@click.argument('pipeline_id', required=True)
def delete_pipeline(pipeline_id):
    """Delete a pipeline

    pipeline_id is required. Returns a string with pipeline id if succeeds, an
    Error string if fails.
    """
    from ploomber import cli as cli_module
    print(cli_module.cloud.delete_pipeline(pipeline_id))


@cloud.command(name='build')
@click.option('-f',
              '--force',
              help='Force execution by ignoring status',
              is_flag=True)
@click.option('--github-number', help="Github's PR number", default=None)
@click.option('--github-owner', help="Github's owner", default=None)
@click.option('--github-repo', help="Github's repo", default=None)
@click.option('--raw', is_flag=True)
def cloud_build(force, github_number, github_owner, github_repo, raw):
    """Build pipeline in the cloud

    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api
    runid = api.upload_project(force,
                               github_number,
                               github_owner,
                               github_repo,
                               verbose=not raw)
    if raw:
        click.echo(runid)


@cloud.command(name="list")
def cloud_list():
    """List cloud executions

    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api
    api.runs()


@cloud.command(name="status")
@click.argument('run_id')
@click.option('--watch', '-w', is_flag=True)
def cloud_status(run_id, watch):
    """Get details on a cloud execution
    $ ploomber cloud status {some-id}


    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api

    if watch:
        idle = 5
        timeout = 10 * 60
        cumsum = 0

        while cumsum < timeout:
            click.clear()
            out = api.run_detail_print(run_id)

            status = set([t['status'] for t in out['tasks']])

            if out['run'] != 'created' and (status == {'finished'}
                                            or 'aborted' in status or 'failed'
                                            in status) or cumsum >= timeout:
                break

            time.sleep(idle)
            cumsum += idle
    else:
        api.run_detail_print(run_id)


@cloud.command(name="products")
@click.option('-d', '--delete', default=None)
def cloud_products(delete):
    """List products in cloud workspace

    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api

    if delete:
        api.delete_products(delete)
    else:
        api.products_list()


@cloud.command(name="download")
@click.argument('pattern')
def cloud_download(pattern):
    """Download products from cloud workspace
    Download all .csv files:
    $ ploomber cloud download '*.csv'


    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api
    api.products_download(pattern)


@cloud.command(name="logs")
@click.argument('run_id')
@click.option('--image', '-i', is_flag=True)
@click.option('--watch', '-w', is_flag=True)
def cloud_logs(run_id, image, watch):
    """Get logs on a cloud execution

    Get task logs:
        $ ploomber cloud logs {some-id}


    Get Docker image building logs:
        $ ploomber cloud logs {some-id} --image

    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api

    if image:
        if watch:
            idle = 2
            timeout = 10 * 60
            cumsum = 0

            while cumsum < timeout:
                click.clear()
                time.sleep(idle)
                api.run_logs_image(run_id, tail=20)
                cumsum += idle
        else:
            api.run_logs_image(run_id)
    else:
        api.run_logs(run_id)


@cloud.command(name="abort")
@click.argument('run_id')
def cloud_abort(run_id):
    """Abort a cloud execution
    $ ploomber cloud abort {some-id}


    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api
    api.run_abort(run_id)


@cloud.command(name="data")
@click.option('-u', '--upload', default=None)
@click.option('-d', '--delete', default=None)
def cloud_data(upload, delete):
    """
    Manage raw data workspace

    List data files:
        $ ploomber cloud data

    Upload data:
        $ ploomber cloud data --upload path/to/data.parquet

    Delete data (deletes all the objects matching the pattern):
        $ ploomber cloud data --delete '*.parquet'

    Currently in private alpha, ask us for an invite:
    https://ploomber.io/community
    """
    from ploomber.cloud import api

    # one arg max

    if upload:
        api.upload_data(upload)
    elif delete:
        api.delete_data(delete)
    else:
        api.data_list()



"""
Handling file I/O for writing files, optionally saving them in chunks
"""
import abc
import warnings
import csv
from pathlib import Path

try:
    import pyarrow as pa
    import pyarrow.parquet as pq
except ImportError:
    pa = None
    pq = None

from ploomber.util import safe_remove, requires


class FileIO(abc.ABC):
    """Abstract class for file I/O
    """
    def __init__(self, path, chunked):
        self.path = Path(path)
        self.chunked = chunked

        # if the file (or if a dir) exists, remove it...
        if self.path.exists():
            safe_remove(self.path)
        # if not...
        else:
            # if parent is currently a file, delete
            if self.path.parent.is_file():
                self.path.parent.unlink()

            # make sure parent exists
            self.path.parent.mkdir(parents=True, exist_ok=True)

        # if this is a chunked file, make sure the file that holds all the
        # chunks exists
        if self.chunked:
            self.path.mkdir()

        # only used when chunked
        self.i = 0

    def write(self, data, headers):
        if self.chunked:
            path = self.path / '{i}.{ext}'.format(i=self.i, ext=self.extension)
            self.write_in_path(str(path), data, headers)
            self.i = self.i + 1
        else:
            self.write_in_path(str(self.path), data, headers)

    @classmethod
    @abc.abstractmethod
    def write_in_path(cls, path, data, headers):
        pass

    @property
    @abc.abstractmethod
    def extension(self):
        pass


class ParquetIO(FileIO):
    """parquet handler

    Notes
    -----
    going from pandas.DataFrame to parquet has an intermediate
    apache arrow conversion (since arrow has the actual implementation
    for writing parquet). pandas provides the pandas.DataFrame.to_parquet
    to do this 2-step process but it gives errors with timestamps:
    ArrowInvalid: Casting from timestamp[ns] to timestamp[ms] would lose data

    This function uses the pyarrow package directly to save to parquet
    """
    @requires(['pyarrow'], 'ParquetIO')
    def __init__(self, path, chunked):
        if chunked:
            warnings.warn('ploomber.io.ParquetIO was initilized with '
                          'chunked=True, '
                          'since each chunk will create a parquet file '
                          'and schema will be inferred chunkwise '
                          'there could be inconsistencies. '
                          'To avoid this, the schema from the first chunk '
                          'will be applied to all chunks, if this causes '
                          'problems, either chunked option off or '
                          'use the ploomber.io.CSVIO')

        super().__init__(path, chunked)
        self.schema = None

    def write(self, data, headers):
        if self.chunked:
            path = self.path / '{i}.{ext}'.format(i=self.i, ext=self.extension)
            schema = self.write_in_path(str(path), data, headers, self.schema)
            self.i = self.i + 1

            if self.i == 0:
                self.schema = schema

                self._logger.info(
                    'Got first chunk, to avoid schemas '
                    'incompatibility, the schema from this chunk '
                    'will be applied to the other chunks, verify '
                    'that this is correct: %s. Columns might be '
                    'incorrectly detected as "null" if all values'
                    ' from the first chunk are empty, in such '
                    'case the only safe way to dump is in one '
                    'chunk (by setting chunksize to None)', schema)
        else:
            self.write_in_path(str(self.path), data, headers, schema=None)

    @classmethod
    def write_in_path(cls, path, data, headers, schema):
        arrays = [pa.array(col) for col in map(list, zip(*data))]
        table = pa.Table.from_arrays(arrays, names=headers, schema=schema)
        pq.write_table(table, str(path))
        return table.schema

    @property
    def extension(self):
        return 'parquet'


class CSVIO(FileIO):
    """csv file handler
    """
    @classmethod
    def write_in_path(cls, path, data, headers):
        with open(path, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile, dialect=csv.unix_dialect)
            writer.writerow(headers)
            writer.writerows(data)

    @property
    def extension(self):
        return 'csv'
